import copy
import dataclasses
import json
import time
import warnings
from asyncio import CancelledError
from typing import Any, AsyncIterator, Protocol, runtime_checkable

import litellm
import structlog
from anthropic import NOT_GIVEN
from anthropic.types.beta.beta_message import BetaMessage as AnthropicMessage
from jinja2 import Template
from litellm.types.router import AllowedFailsPolicy
from litellm.utils import CustomStreamWrapper, ModelResponse
from openai import AsyncOpenAI
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk
from pydantic import BaseModel

from skyvern.config import settings
from skyvern.exceptions import SkyvernContextWindowExceededError
from skyvern.forge import app
from skyvern.forge.forge_openai_client import ForgeAsyncHttpxClientWrapper
from skyvern.forge.sdk.api.llm.api_handler import LLMAPIHandler, dummy_llm_api_handler
from skyvern.forge.sdk.api.llm.config_registry import LLMConfigRegistry
from skyvern.forge.sdk.api.llm.exceptions import (
    DuplicateCustomLLMProviderError,
    InvalidLLMConfigError,
    LLMProviderError,
    LLMProviderErrorRetryableTask,
)
from skyvern.forge.sdk.api.llm.models import (
    LLMAllowedFailsPolicy,
    LLMConfig,
    LLMRouterConfig,
)
from skyvern.forge.sdk.api.llm.ui_tars_response import UITarsResponse
from skyvern.forge.sdk.api.llm.utils import (
    is_image_message,
    llm_messages_builder,
    llm_messages_builder_with_history,
    parse_api_response,
)
from skyvern.forge.sdk.artifact.manager import BulkArtifactCreationRequest
from skyvern.forge.sdk.artifact.models import ArtifactType
from skyvern.forge.sdk.core import skyvern_context
from skyvern.forge.sdk.core.skyvern_context import SkyvernContext
from skyvern.forge.sdk.models import SpeculativeLLMMetadata, Step
from skyvern.forge.sdk.schemas.ai_suggestions import AISuggestion
from skyvern.forge.sdk.schemas.task_v2 import TaskV2, Thought
from skyvern.forge.sdk.trace import traced
from skyvern.utils.image_resizer import Resolution, get_resize_target_dimension, resize_screenshots

LOG = structlog.get_logger()

EXTRACT_ACTION_PROMPT_NAME = "extract-actions"
CHECK_USER_GOAL_PROMPT_NAMES = {"check-user-goal", "check-user-goal-with-termination"}

# Default thinking budgets (configurable via env vars, can be overridden by THINKING_BUDGET_OPTIMIZATION experiment)
EXTRACT_ACTION_DEFAULT_THINKING_BUDGET = settings.EXTRACT_ACTION_THINKING_BUDGET
DEFAULT_THINKING_BUDGET = settings.DEFAULT_THINKING_BUDGET


def _safe_model_dump_json(response: ModelResponse, indent: int = 2) -> str:
    """
    Call model_dump_json() while suppressing Pydantic serialization warnings.

    litellm's ModelResponse has Union types (Choices | StreamingChoices) that cause
    Pydantic to emit warnings during serialization. These are cosmetic and don't
    affect the serialized output.
    """
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning, message=".*Pydantic.*Serializ.*")
        return response.model_dump_json(indent=indent)


@runtime_checkable
class RouterWithModelList(Protocol):
    model_list: list[dict[str, Any]]


def _get_primary_model_dict(router: Any, main_model_group: str) -> dict[str, Any] | None:
    if isinstance(router, RouterWithModelList):
        for model_dict in router.model_list:
            if model_dict.get("model_name") == main_model_group:
                return model_dict
    return None


class LLMCallStats(BaseModel):
    input_tokens: int | None = None
    output_tokens: int | None = None
    reasoning_tokens: int | None = None
    cached_tokens: int | None = None
    llm_cost: float | None = None


def _get_artifact_targets_and_persist_flag(
    step: Step | None,
    is_speculative_step: bool,
    task_v2: TaskV2 | None,
    thought: Thought | None,
    ai_suggestion: AISuggestion | None,
) -> tuple[bool, dict[str, Any]]:
    artifact_targets = {
        "step": step if not is_speculative_step else None,
        "task_v2": task_v2,
        "thought": thought,
        "ai_suggestion": ai_suggestion,
    }
    has_artifact_target = any(value is not None for value in artifact_targets.values())
    should_persist_llm_artifacts = not is_speculative_step and has_artifact_target
    return should_persist_llm_artifacts, artifact_targets


async def _log_hashed_href_map_artifacts_if_needed(
    artifacts: list[BulkArtifactCreationRequest | None],
    context: SkyvernContext | None,
    step: Step | None,
    task_v2: TaskV2 | None,
    thought: Thought | None,
    ai_suggestion: AISuggestion | None,
    *,
    is_speculative_step: bool,
) -> None:
    should_persist_llm_artifacts, artifact_targets = _get_artifact_targets_and_persist_flag(
        step, is_speculative_step, task_v2, thought, ai_suggestion
    )
    if context and context.hashed_href_map and should_persist_llm_artifacts:
        artifacts.append(
            await app.ARTIFACT_MANAGER.prepare_llm_artifact(
                data=json.dumps(context.hashed_href_map, indent=2).encode("utf-8"),
                artifact_type=ArtifactType.HASHED_HREF_MAP,
                **artifact_targets,
            )
        )


def _log_vertex_cache_hit_if_needed(
    context: SkyvernContext | None,
    prompt_name: str,
    llm_identifier: str,
    cached_tokens: int,
) -> None:
    if cached_tokens > 0 and prompt_name == EXTRACT_ACTION_PROMPT_NAME and context and context.vertex_cache_name:
        LOG.info(
            "Vertex cache hit",
            prompt_name=prompt_name,
            llm_key=llm_identifier,
            cached_tokens=cached_tokens,
            cache_name=context.vertex_cache_name,
            cache_key=context.vertex_cache_key,
            cache_variant=context.vertex_cache_variant,
        )


def _convert_allowed_fails_policy(policy: LLMAllowedFailsPolicy | None) -> AllowedFailsPolicy | None:
    if policy is None:
        return None

    return AllowedFailsPolicy(
        BadRequestErrorAllowedFails=policy.bad_request_error_allowed_fails,
        AuthenticationErrorAllowedFails=policy.authentication_error_allowed_fails,
        TimeoutErrorAllowedFails=policy.timeout_error_allowed_fails,
        RateLimitErrorAllowedFails=policy.rate_limit_error_allowed_fails,
        ContentPolicyViolationErrorAllowedFails=policy.content_policy_violation_error_allowed_fails,
        InternalServerErrorAllowedFails=policy.internal_server_error_allowed_fails,
    )


class LLMAPIHandlerFactory:
    _custom_handlers: dict[str, LLMAPIHandler] = {}
    _router_handler_cache: dict[str, LLMAPIHandler] = {}
    _thinking_budget_settings: dict[str, int] | None = None
    _prompt_caching_settings: dict[str, bool] | None = None

    @staticmethod
    def _strip_static_prompt_from_messages(messages: list[dict[str, Any]], static_prompt: str) -> bool:
        """
        Strips the static prompt from the first matching user message in the list.
        Returns True if the prompt was found and stripped, False otherwise.

        This handles both string content and list-based content (e.g. for vision models).
        The static prompt is right-stripped to handle trailing newlines from templates.
        The remaining dynamic content is left-stripped to handle connector whitespace.
        """
        static_text = static_prompt.rstrip()
        prompt_stripped = False

        for msg in messages:
            if msg.get("role") == "user":
                content = msg.get("content")
                if isinstance(content, str):
                    if content.startswith(static_text):
                        msg["content"] = content[len(static_text) :].lstrip()
                        prompt_stripped = True
                        break
                elif isinstance(content, list):
                    for block in content:
                        if block.get("type") == "text":
                            text = block.get("text", "")
                            if text.startswith(static_text):
                                block["text"] = text[len(static_text) :].lstrip()
                                prompt_stripped = True
                                break
                    if prompt_stripped:
                        break

        return prompt_stripped

    @staticmethod
    def _models_equivalent(left: str | None, right: str | None) -> bool:
        """Used only by `llm_api_handler_with_router_and_fallback`. Router model
        groups carry the `vertex-` prefix while LiteLLM responses return the
        underlying provider label (e.g. `gemini-2.5-pro`). Stripping the prefix
        lets us detect whether the configured primary (the router's
        `main_model_group`) actually served the request without replumbing every
        config/registry reference.
        """
        if left == right:
            return True
        if left is None or right is None:
            return False

        def _normalize(label: str) -> str:
            normalized = label.lower()
            return normalized[len("vertex-") :] if normalized.startswith("vertex-") else normalized

        return _normalize(left) == _normalize(right)

    @staticmethod
    def _extract_token_counts(response: ModelResponse | CustomStreamWrapper) -> tuple[int, int, int, int]:
        """Extract token counts from response usage information."""
        input_tokens = 0
        output_tokens = 0
        reasoning_tokens = 0
        cached_tokens = 0

        if hasattr(response, "usage") and response.usage:
            input_tokens = getattr(response.usage, "prompt_tokens", 0)
            output_tokens = getattr(response.usage, "completion_tokens", 0)

            # Extract reasoning tokens from completion_tokens_details
            completion_token_detail = getattr(response.usage, "completion_tokens_details", None)
            if completion_token_detail:
                reasoning_tokens = getattr(completion_token_detail, "reasoning_tokens", 0) or 0

            # Extract cached tokens from prompt_tokens_details
            cached_token_detail = getattr(response.usage, "prompt_tokens_details", None)
            if cached_token_detail:
                cached_tokens = getattr(cached_token_detail, "cached_tokens", 0) or 0

            # Fallback: Some providers expose cache_read_input_tokens directly on usage
            if cached_tokens == 0:
                cached_tokens = getattr(response.usage, "cache_read_input_tokens", 0) or 0

        return input_tokens, output_tokens, reasoning_tokens, cached_tokens

    @staticmethod
    def _apply_thinking_budget_optimization(
        parameters: dict[str, Any], new_budget: int, llm_config: LLMConfig | LLMRouterConfig, prompt_name: str
    ) -> None:
        """Apply thinking budget optimization based on model type and LiteLLM reasoning support."""
        # Compute a safe model label and a representative model for capability checks
        model_label = LLMAPIHandlerFactory._get_model_label(llm_config)
        check_model = LLMAPIHandlerFactory._get_check_model(llm_config, model_label)

        # Check reasoning support (safe call - log but don't fail if litellm errors)
        supports_reasoning = False
        if check_model:
            try:
                supports_reasoning = litellm.supports_reasoning(model=check_model)
            except Exception as exc:  # pragma: no cover - diagnostic safeguard
                LOG.debug(
                    "Failed to check reasoning support via litellm",
                    model=check_model,
                    error=str(exc),
                )

        try:
            # Gemini router/fallback configs (e.g., gemini-2.5-pro-gpt-5-fallback-router)
            # are not recognized by litellm, but they do support reasoning budgets.
            if not supports_reasoning and model_label:
                model_label_lower = model_label.lower()
                if "gemini" in model_label_lower and "fallback" in model_label_lower:
                    supports_reasoning = True
                    LOG.info(
                        "Forcing reasoning support for Gemini fallback model",
                        prompt_name=prompt_name,
                        budget=new_budget,
                        model=model_label,
                    )

            if check_model and not supports_reasoning:
                LOG.info(
                    "Thinking budget optimization not supported for model",
                    prompt_name=prompt_name,
                    budget=new_budget,
                    model=model_label,
                )
                return
            # Apply optimization based on model type
            model_label_lower = (model_label or "").lower()
            if "gemini" in model_label_lower:
                # Gemini models use the exact integer budget value
                LLMAPIHandlerFactory._apply_gemini_thinking_optimization(
                    parameters, new_budget, llm_config, prompt_name
                )
            elif "anthropic" in model_label_lower or "claude" in model_label_lower:
                # Anthropic/Claude models use "low" for all budget values (per LiteLLM constants)
                LLMAPIHandlerFactory._apply_anthropic_thinking_optimization(
                    parameters, new_budget, llm_config, prompt_name
                )
            else:
                # Other reasoning-capable models (Deepseek, etc.) - use "low" for all budget values
                parameters["reasoning_effort"] = "low"
                LOG.debug(
                    "Applied thinking budget optimization (reasoning_effort)",
                    prompt_name=prompt_name,
                    budget=new_budget,
                    reasoning_effort="low",
                    model=model_label,
                )
        except (AttributeError, KeyError, TypeError) as e:
            LOG.warning(
                "Failed to apply thinking budget optimization",
                prompt_name=prompt_name,
                budget=new_budget,
                model=model_label,
                error=str(e),
                exc_info=True,
            )

    @staticmethod
    def _apply_anthropic_thinking_optimization(
        parameters: dict[str, Any], new_budget: int, llm_config: LLMConfig | LLMRouterConfig, prompt_name: str
    ) -> None:
        """Apply thinking optimization for Anthropic/Claude models."""
        model_label = LLMAPIHandlerFactory._get_model_label(llm_config)

        if llm_config.reasoning_effort is not None:
            # Use reasoning_effort if configured in LLM config - always use "low" per LiteLLM constants
            parameters["reasoning_effort"] = "low"
            # Get safe model label for logging
            model_label = getattr(llm_config, "model_name", None)
            if model_label is None and isinstance(llm_config, LLMRouterConfig):
                model_label = getattr(llm_config, "main_model_group", "router")

            LOG.debug(
                "Applied thinking budget optimization (reasoning_effort)",
                prompt_name=prompt_name,
                budget=new_budget,
                reasoning_effort="low",
                model=model_label,
            )
        else:
            # Use thinking parameter with budget_tokens for Anthropic models
            if "thinking" in parameters and isinstance(parameters["thinking"], dict):
                parameters["thinking"]["budget_tokens"] = new_budget
            else:
                parameters["thinking"] = {"budget_tokens": new_budget, "type": "enabled"}
            # Get safe model label for logging
            model_label = getattr(llm_config, "model_name", None)
            if model_label is None and isinstance(llm_config, LLMRouterConfig):
                model_label = getattr(llm_config, "main_model_group", "router")

            LOG.debug(
                "Applied thinking budget optimization (thinking)",
                prompt_name=prompt_name,
                budget=new_budget,
                model=model_label,
            )

    @staticmethod
    def _apply_gemini_thinking_optimization(
        parameters: dict[str, Any], new_budget: int, llm_config: LLMConfig | LLMRouterConfig, prompt_name: str
    ) -> None:
        """Apply thinking optimization for Gemini models using exact integer budget value."""
        model_label = LLMAPIHandlerFactory._get_model_label(llm_config)

        # Models that use thinking_level (e.g. Gemini 3 Pro/Flash) don't support budget_tokens.
        # Their reasoning is already bounded by the thinking_level set in their config, so skip.
        if "thinking_level" in parameters:
            return

        if "thinking" in parameters and isinstance(parameters["thinking"], dict):
            parameters["thinking"]["budget_tokens"] = new_budget
        else:
            thinking_payload: dict[str, Any] = {"budget_tokens": new_budget}
            if settings.GEMINI_INCLUDE_THOUGHT:
                thinking_payload["type"] = "enabled"
            parameters["thinking"] = thinking_payload

        LOG.debug(
            "Applied thinking budget optimization (budget_tokens)",
            prompt_name=prompt_name,
            budget=new_budget,
            model=model_label,
        )

    @staticmethod
    def _get_model_label(llm_config: LLMConfig | LLMRouterConfig) -> str | None:
        """Extract a safe model label from LLMConfig or LLMRouterConfig for logging and capability checks."""
        model_label = getattr(llm_config, "model_name", None)
        # Compute a safe model label and a representative model for capability checks
        if model_label is None and isinstance(llm_config, LLMRouterConfig):
            model_label = getattr(llm_config, "main_model_group", "router")
        return model_label

    @staticmethod
    def _get_check_model(llm_config: LLMConfig | LLMRouterConfig, model_label: str | None) -> str | None:
        """Get a representative model name for capability checks from LLMRouterConfig or use the model label."""
        check_model = model_label
        if isinstance(llm_config, LLMRouterConfig) and getattr(llm_config, "model_list", None):
            try:
                check_model = llm_config.model_list[0].model_name or model_label  # type: ignore[attr-defined]
            except Exception:
                check_model = model_label
        return check_model

    @staticmethod
    def is_github_copilot_endpoint() -> bool:
        """Check if the OPENAI_COMPATIBLE endpoint is GitHub Copilot."""
        return (
            settings.OPENAI_COMPATIBLE_API_BASE is not None
            and settings.OPENAI_COMPATIBLE_GITHUB_COPILOT_DOMAIN in settings.OPENAI_COMPATIBLE_API_BASE
        )

    @staticmethod
    def get_override_llm_api_handler(override_llm_key: str | None, *, default: LLMAPIHandler) -> LLMAPIHandler:
        if not override_llm_key:
            return default
        try:
            # Explicit overrides should honor the exact model choice and skip experimentation reroutes.
            return LLMAPIHandlerFactory.get_llm_api_handler(override_llm_key)
        except Exception:
            LOG.warning(
                "Failed to get override LLM API handler, going to use the default.",
                override_llm_key=override_llm_key,
                exc_info=True,
            )
            return default

    @staticmethod
    def get_llm_api_handler_with_router(llm_key: str) -> LLMAPIHandler:
        if llm_key in LLMAPIHandlerFactory._router_handler_cache:
            return LLMAPIHandlerFactory._router_handler_cache[llm_key]

        llm_config = LLMConfigRegistry.get_config(llm_key)
        if not isinstance(llm_config, LLMRouterConfig):
            raise InvalidLLMConfigError(llm_key)

        router = litellm.Router(
            model_list=[dataclasses.asdict(model) for model in llm_config.model_list],
            redis_host=llm_config.redis_host,
            redis_port=llm_config.redis_port,
            redis_password=llm_config.redis_password,
            routing_strategy=llm_config.routing_strategy,
            fallbacks=(
                [{llm_config.main_model_group: [llm_config.fallback_model_group]}]
                if llm_config.fallback_model_group
                else []
            ),
            num_retries=llm_config.num_retries,
            retry_after=llm_config.retry_delay_seconds,
            disable_cooldowns=llm_config.disable_cooldowns,
            allowed_fails=llm_config.allowed_fails,
            allowed_fails_policy=_convert_allowed_fails_policy(llm_config.allowed_fails_policy),
            cooldown_time=llm_config.cooldown_time,
            set_verbose=(False if settings.is_cloud_environment() else llm_config.set_verbose),
            enable_pre_call_checks=True,
        )
        main_model_group = llm_config.main_model_group

        @traced(tags=[llm_key])
        async def llm_api_handler_with_router_and_fallback(
            prompt: str,
            prompt_name: str,
            step: Step | None = None,
            task_v2: TaskV2 | None = None,
            thought: Thought | None = None,
            ai_suggestion: AISuggestion | None = None,
            screenshots: list[bytes] | None = None,
            parameters: dict[str, Any] | None = None,
            organization_id: str | None = None,
            tools: list | None = None,
            use_message_history: bool = False,
            raw_response: bool = False,
            window_dimension: Resolution | None = None,
            force_dict: bool = True,
        ) -> dict[str, Any] | Any:
            """
            Custom LLM API handler that utilizes the LiteLLM router and fallbacks to OpenAI GPT-4 Vision.

            Args:
                prompt: The prompt to generate completions for.
                step: The step object associated with the prompt.
                screenshots: The screenshots associated with the prompt.
                parameters: Additional parameters to be passed to the LLM router.

            Returns:
                The response from the LLM router.
            """
            start_time = time.time()

            if parameters is None:
                parameters = LLMAPIHandlerFactory.get_api_parameters(llm_config)

            # Apply thinking budget optimization if settings are available
            if (
                LLMAPIHandlerFactory._thinking_budget_settings
                and prompt_name in LLMAPIHandlerFactory._thinking_budget_settings
            ):
                new_budget = LLMAPIHandlerFactory._thinking_budget_settings[prompt_name]
                LLMAPIHandlerFactory._apply_thinking_budget_optimization(
                    parameters, new_budget, llm_config, prompt_name
                )
            elif prompt_name == EXTRACT_ACTION_PROMPT_NAME:
                # Apply default thinking budget for extract-actions (512) unless overridden by experiment
                LLMAPIHandlerFactory._apply_thinking_budget_optimization(
                    parameters, EXTRACT_ACTION_DEFAULT_THINKING_BUDGET, llm_config, prompt_name
                )
            else:
                # Apply default thinking budget for all other prompts to prevent unbounded reasoning
                LLMAPIHandlerFactory._apply_thinking_budget_optimization(
                    parameters, DEFAULT_THINKING_BUDGET, llm_config, prompt_name
                )

            context = skyvern_context.current()
            is_speculative_step = step.is_speculative if step else False
            should_persist_llm_artifacts, artifact_targets = _get_artifact_targets_and_persist_flag(
                step, is_speculative_step, task_v2, thought, ai_suggestion
            )

            artifacts: list[BulkArtifactCreationRequest | None] = []
            try:
                await _log_hashed_href_map_artifacts_if_needed(
                    artifacts,
                    context,
                    step,
                    task_v2,
                    thought,
                    ai_suggestion,
                    is_speculative_step=is_speculative_step,
                )

                llm_prompt_value = prompt
                if should_persist_llm_artifacts:
                    artifacts.append(
                        await app.ARTIFACT_MANAGER.prepare_llm_artifact(
                            data=llm_prompt_value.encode("utf-8"),
                            artifact_type=ArtifactType.LLM_PROMPT,
                            screenshots=screenshots,
                            **artifact_targets,
                        )
                    )
                # Build messages and apply caching in one step
                messages = await llm_messages_builder(prompt, screenshots, llm_config.add_assistant_prefix)

                async def _log_llm_request_artifact(model_label: str, vertex_cache_attached_flag: bool) -> str:
                    llm_request_payload = {
                        "model": model_label,
                        "messages": messages,
                        **parameters,
                        "vertex_cache_attached": vertex_cache_attached_flag,
                    }
                    llm_request_json = json.dumps(llm_request_payload)
                    if should_persist_llm_artifacts:
                        artifacts.append(
                            await app.ARTIFACT_MANAGER.prepare_llm_artifact(
                                data=llm_request_json.encode("utf-8"),
                                artifact_type=ArtifactType.LLM_REQUEST,
                                **artifact_targets,
                            )
                        )
                    return llm_request_json

                # Inject context caching system message when available
                # IMPORTANT: Only inject for extract-actions prompt to avoid contaminating other prompts
                # (e.g., check-user-goal) with the extract-action schema
                try:
                    if (
                        context
                        and context.cached_static_prompt
                        and prompt_name == EXTRACT_ACTION_PROMPT_NAME  # Only inject for extract-actions
                        and isinstance(llm_config, LLMConfig)
                        and isinstance(llm_config.model_name, str)
                    ):
                        # Check if this is an OpenAI model
                        if (
                            llm_config.model_name.startswith("gpt-")
                            or llm_config.model_name.startswith("o1-")
                            or llm_config.model_name.startswith("o3-")
                        ):
                            # For OpenAI models, we need to add the cached content as a system message
                            # and mark it for caching using the cache_control parameter
                            caching_system_message = {
                                "role": "system",
                                "content": [
                                    {
                                        "type": "text",
                                        "text": context.cached_static_prompt,
                                    }
                                ],
                            }
                            messages = [caching_system_message] + messages
                            LOG.info(
                                "Applied OpenAI context caching",
                                prompt_name=prompt_name,
                                model=llm_config.model_name,
                            )
                except Exception as e:
                    LOG.warning("Failed to apply context caching system message", error=str(e), exc_info=True)

                cache_resource_name = getattr(context, "vertex_cache_name", None)
                cache_variant = getattr(context, "vertex_cache_variant", None)
                primary_model_dict = _get_primary_model_dict(router, main_model_group)
                should_attach_vertex_cache = bool(
                    cache_resource_name is not None
                    and prompt_name == EXTRACT_ACTION_PROMPT_NAME
                    and getattr(context, "use_prompt_caching", False)
                    and main_model_group
                    and "gemini" in main_model_group.lower()
                    and primary_model_dict is not None
                )

                model_used = main_model_group
                llm_request_json = ""

                async def _call_primary_with_vertex_cache(
                    cache_name: str,
                    cache_variant_name: str | None,
                ) -> tuple[ModelResponse, str, str]:
                    if primary_model_dict is None:
                        raise ValueError("Primary router model missing configuration")
                    litellm_params = copy.deepcopy(primary_model_dict.get("litellm_params") or {})
                    if not litellm_params:
                        raise ValueError("Primary router model missing litellm_params")
                    active_params = copy.deepcopy(litellm_params)
                    active_params.update(parameters)
                    active_params["cached_content"] = cache_name
                    request_model = active_params.pop("model", primary_model_dict.get("model_name", main_model_group))

                    # Clone messages to avoid modifying original list which is needed for fallback
                    active_messages = copy.deepcopy(messages)

                    # Strip static prompt from the request messages because it's already in the cache
                    # Sending it again causes double-billing (once cached, once uncached)
                    if context and context.cached_static_prompt:
                        prompt_stripped = LLMAPIHandlerFactory._strip_static_prompt_from_messages(
                            active_messages, context.cached_static_prompt
                        )

                        if prompt_stripped:
                            LOG.info("Stripped static prompt from cached request to avoid double-billing")
                        else:
                            LOG.warning("Could not find static prompt to strip from cached request")

                    LOG.info(
                        "Adding Vertex AI cache reference to primary Gemini request",
                        prompt_name=prompt_name,
                        primary_model=main_model_group,
                        fallback_model=llm_config.fallback_model_group,
                        cache_name=cache_name,
                        cache_key=getattr(context, "vertex_cache_key", None),
                        cache_variant=cache_variant_name,
                    )
                    request_payload_json = await _log_llm_request_artifact(request_model, True)
                    response = await litellm.acompletion(
                        model=request_model,
                        messages=active_messages,
                        timeout=settings.LLM_CONFIG_TIMEOUT,
                        drop_params=True,
                        **active_params,
                    )
                    return response, request_model, request_payload_json

                async def _call_router_without_cache() -> tuple[ModelResponse, str]:
                    request_payload_json = await _log_llm_request_artifact(llm_key, False)
                    response = await router.acompletion(
                        model=main_model_group,
                        messages=messages,
                        timeout=settings.LLM_CONFIG_TIMEOUT,
                        drop_params=True,
                        **parameters,
                    )
                    return response, request_payload_json

                try:
                    response: ModelResponse | None = None
                    if should_attach_vertex_cache and cache_resource_name:
                        try:
                            response, direct_model_used, llm_request_json = await _call_primary_with_vertex_cache(
                                cache_resource_name,
                                cache_variant,
                            )
                            model_used = response.model or direct_model_used
                        except CancelledError:
                            raise
                        except Exception as cache_error:
                            LOG.warning(
                                "Vertex cache primary call failed, retrying via router",
                                prompt_name=prompt_name,
                                error=str(cache_error),
                                cache_name=cache_resource_name,
                                cache_variant=cache_variant,
                            )
                            response = None

                    if response is None:
                        response, llm_request_json = await _call_router_without_cache()
                        response_model = response.model or main_model_group
                        model_used = response_model
                        if not LLMAPIHandlerFactory._models_equivalent(response_model, main_model_group):
                            LOG.info(
                                "LLM router fallback succeeded",
                                llm_key=llm_key,
                                prompt_name=prompt_name,
                                primary_model=main_model_group,
                                fallback_model=response_model,
                            )
                except litellm.exceptions.APIError as e:
                    raise LLMProviderErrorRetryableTask(llm_key) from e
                except litellm.exceptions.ContextWindowExceededError as e:
                    duration_seconds = time.time() - start_time
                    LOG.exception(
                        "Context window exceeded",
                        llm_key=llm_key,
                        model=main_model_group,
                        prompt_name=prompt_name,
                        duration_seconds=duration_seconds,
                    )
                    raise SkyvernContextWindowExceededError() from e
                except ValueError as e:
                    duration_seconds = time.time() - start_time
                    LOG.exception(
                        "LLM token limit exceeded",
                        llm_key=llm_key,
                        model=main_model_group,
                        prompt_name=prompt_name,
                        duration_seconds=duration_seconds,
                    )
                    raise LLMProviderErrorRetryableTask(llm_key) from e
                except Exception as e:
                    duration_seconds = time.time() - start_time
                    LOG.exception(
                        "LLM request failed unexpectedly",
                        llm_key=llm_key,
                        model=main_model_group,
                        prompt_name=prompt_name,
                        duration_seconds=duration_seconds,
                    )
                    raise LLMProviderError(llm_key) from e

                llm_response_json = _safe_model_dump_json(response)
                if should_persist_llm_artifacts:
                    artifacts.append(
                        await app.ARTIFACT_MANAGER.prepare_llm_artifact(
                            data=llm_response_json.encode("utf-8"),
                            artifact_type=ArtifactType.LLM_RESPONSE,
                            **artifact_targets,
                        )
                    )
                prompt_tokens = 0
                completion_tokens = 0
                reasoning_tokens = 0
                cached_tokens = 0
                completion_token_detail = None
                cached_token_detail = None
                try:
                    # FIXME: volcengine doesn't support litellm cost calculation.
                    llm_cost = litellm.completion_cost(completion_response=response)
                except Exception as e:
                    LOG.debug("Failed to calculate LLM cost", error=str(e), exc_info=True)
                    llm_cost = 0
                prompt_tokens = 0
                completion_tokens = 0
                reasoning_tokens = 0
                cached_tokens = 0

                if hasattr(response, "usage") and response.usage:
                    prompt_tokens = getattr(response.usage, "prompt_tokens", 0)
                    completion_tokens = getattr(response.usage, "completion_tokens", 0)

                    # Extract reasoning tokens from completion_tokens_details
                    completion_token_detail = getattr(response.usage, "completion_tokens_details", None)
                    if completion_token_detail:
                        reasoning_tokens = getattr(completion_token_detail, "reasoning_tokens", 0) or 0

                    # Extract cached tokens from prompt_tokens_details
                    cached_token_detail = getattr(response.usage, "prompt_tokens_details", None)
                    if cached_token_detail:
                        cached_tokens = getattr(cached_token_detail, "cached_tokens", 0) or 0

                    # Fallback for Vertex/Gemini: LiteLLM exposes cache_read_input_tokens on usage
                    if cached_tokens == 0:
                        cached_tokens = getattr(response.usage, "cache_read_input_tokens", 0) or 0
                if step and not is_speculative_step:
                    await app.DATABASE.update_step(
                        task_id=step.task_id,
                        step_id=step.step_id,
                        organization_id=step.organization_id,
                        incremental_cost=llm_cost,
                        incremental_input_tokens=prompt_tokens if prompt_tokens > 0 else None,
                        incremental_output_tokens=completion_tokens if completion_tokens > 0 else None,
                        incremental_reasoning_tokens=reasoning_tokens if reasoning_tokens > 0 else None,
                        incremental_cached_tokens=cached_tokens if cached_tokens > 0 else None,
                    )
                if thought:
                    await app.DATABASE.update_thought(
                        thought_id=thought.observer_thought_id,
                        organization_id=thought.organization_id,
                        input_token_count=prompt_tokens if prompt_tokens > 0 else None,
                        output_token_count=completion_tokens if completion_tokens > 0 else None,
                        thought_cost=llm_cost,
                        reasoning_token_count=reasoning_tokens if reasoning_tokens > 0 else None,
                        cached_token_count=cached_tokens if cached_tokens > 0 else None,
                    )
                parsed_response = parse_api_response(response, llm_config.add_assistant_prefix, force_dict)
                parsed_response_json = json.dumps(parsed_response, indent=2)
                if should_persist_llm_artifacts:
                    artifacts.append(
                        await app.ARTIFACT_MANAGER.prepare_llm_artifact(
                            data=parsed_response_json.encode("utf-8"),
                            artifact_type=ArtifactType.LLM_RESPONSE_PARSED,
                            **artifact_targets,
                        )
                    )

                rendered_response_json = None
                if context and len(context.hashed_href_map) > 0:
                    llm_content = json.dumps(parsed_response)
                    rendered_content = Template(llm_content).render(context.hashed_href_map)
                    parsed_response = json.loads(rendered_content)
                    rendered_response_json = json.dumps(parsed_response, indent=2)
                    if should_persist_llm_artifacts:
                        artifacts.append(
                            await app.ARTIFACT_MANAGER.prepare_llm_artifact(
                                data=rendered_response_json.encode("utf-8"),
                                artifact_type=ArtifactType.LLM_RESPONSE_RENDERED,
                                **artifact_targets,
                            )
                        )

                # Track LLM API handler duration, token counts, and cost
                organization_id = organization_id or (
                    step.organization_id if step else (thought.organization_id if thought else None)
                )
                duration_seconds = time.time() - start_time
                LOG.info(
                    "LLM API handler duration metrics",
                    llm_key=llm_key,
                    model=model_used,
                    prompt_name=prompt_name,
                    duration_seconds=duration_seconds,
                    step_id=step.step_id if step else None,
                    thought_id=thought.observer_thought_id if thought else None,
                    organization_id=organization_id,
                    input_tokens=prompt_tokens if prompt_tokens > 0 else None,
                    output_tokens=completion_tokens if completion_tokens > 0 else None,
                    reasoning_tokens=reasoning_tokens if reasoning_tokens > 0 else None,
                    cached_tokens=cached_tokens if cached_tokens > 0 else None,
                    llm_cost=llm_cost if llm_cost > 0 else None,
                )

                if step and is_speculative_step:
                    step.speculative_llm_metadata = SpeculativeLLMMetadata(
                        prompt=llm_prompt_value,
                        llm_request_json=llm_request_json,
                        llm_response_json=llm_response_json,
                        parsed_response_json=parsed_response_json,
                        rendered_response_json=rendered_response_json,
                        llm_key=llm_key,
                        model=model_used,
                        duration_seconds=duration_seconds,
                        input_tokens=prompt_tokens if prompt_tokens > 0 else None,
                        output_tokens=completion_tokens if completion_tokens > 0 else None,
                        reasoning_tokens=reasoning_tokens if reasoning_tokens > 0 else None,
                        cached_tokens=cached_tokens if cached_tokens > 0 else None,
                        llm_cost=llm_cost if llm_cost > 0 else None,
                    )

                return parsed_response
            finally:
                try:
                    await app.ARTIFACT_MANAGER.bulk_create_artifacts(artifacts)
                except Exception:
                    LOG.error("Failed to persist artifacts", exc_info=True)

        llm_api_handler_with_router_and_fallback.llm_key = llm_key  # type: ignore[attr-defined]
        LLMAPIHandlerFactory._router_handler_cache[llm_key] = llm_api_handler_with_router_and_fallback
        return llm_api_handler_with_router_and_fallback

    @staticmethod
    def get_llm_api_handler(
        llm_key: str,
        base_parameters: dict[str, Any] | None = None,
    ) -> LLMAPIHandler:
        try:
            llm_config = LLMConfigRegistry.get_config(llm_key)
        except InvalidLLMConfigError:
            return dummy_llm_api_handler

        if LLMConfigRegistry.is_router_config(llm_key):
            return LLMAPIHandlerFactory.get_llm_api_handler_with_router(llm_key)

        # For OpenRouter models, use LLMCaller which has native OpenRouter support
        if llm_key.startswith("openrouter/"):
            llm_caller = LLMCaller(llm_key=llm_key, base_parameters=base_parameters)
            return llm_caller.call

        # For GitHub Copilot via OPENAI_COMPATIBLE, use LLMCaller for a custom header
        if llm_key == "OPENAI_COMPATIBLE" and LLMAPIHandlerFactory.is_github_copilot_endpoint():
            llm_caller = LLMCaller(llm_key=llm_key, base_parameters=base_parameters)
            return llm_caller.call

        assert isinstance(llm_config, LLMConfig)

        @traced(tags=[llm_key])
        async def llm_api_handler(
            prompt: str,
            prompt_name: str,
            step: Step | None = None,
            task_v2: TaskV2 | None = None,
            thought: Thought | None = None,
            ai_suggestion: AISuggestion | None = None,
            screenshots: list[bytes] | None = None,
            parameters: dict[str, Any] | None = None,
            organization_id: str | None = None,
            tools: list | None = None,
            use_message_history: bool = False,
            raw_response: bool = False,
            window_dimension: Resolution | None = None,
            force_dict: bool = True,
        ) -> dict[str, Any] | Any:
            start_time = time.time()
            active_parameters = base_parameters or {}
            if parameters is None:
                parameters = LLMAPIHandlerFactory.get_api_parameters(llm_config)

            active_parameters.update(parameters)
            if llm_config.litellm_params:  # type: ignore
                active_parameters.update(llm_config.litellm_params)  # type: ignore

            if "timeout" not in active_parameters:
                active_parameters["timeout"] = settings.LLM_CONFIG_TIMEOUT

            # Apply thinking budget optimization if settings are available
            if (
                LLMAPIHandlerFactory._thinking_budget_settings
                and prompt_name in LLMAPIHandlerFactory._thinking_budget_settings
            ):
                new_budget = LLMAPIHandlerFactory._thinking_budget_settings[prompt_name]
                LLMAPIHandlerFactory._apply_thinking_budget_optimization(
                    active_parameters, new_budget, llm_config, prompt_name
                )
            elif prompt_name == EXTRACT_ACTION_PROMPT_NAME:
                # Apply default thinking budget for extract-actions (512) unless overridden by experiment
                LLMAPIHandlerFactory._apply_thinking_budget_optimization(
                    active_parameters, EXTRACT_ACTION_DEFAULT_THINKING_BUDGET, llm_config, prompt_name
                )
            else:
                # Apply default thinking budget for all other prompts to prevent unbounded reasoning
                LLMAPIHandlerFactory._apply_thinking_budget_optimization(
                    active_parameters, DEFAULT_THINKING_BUDGET, llm_config, prompt_name
                )

            context = skyvern_context.current()
            is_speculative_step = step.is_speculative if step else False
            should_persist_llm_artifacts, artifact_targets = _get_artifact_targets_and_persist_flag(
                step, is_speculative_step, task_v2, thought, ai_suggestion
            )

            artifacts: list[BulkArtifactCreationRequest | None] = []
            try:
                await _log_hashed_href_map_artifacts_if_needed(
                    artifacts,
                    context,
                    step,
                    task_v2,
                    thought,
                    ai_suggestion,
                    is_speculative_step=is_speculative_step,
                )

                llm_prompt_value = prompt
                if should_persist_llm_artifacts:
                    artifacts.append(
                        await app.ARTIFACT_MANAGER.prepare_llm_artifact(
                            data=llm_prompt_value.encode("utf-8"),
                            artifact_type=ArtifactType.LLM_PROMPT,
                            screenshots=screenshots,
                            **artifact_targets,
                        )
                    )

                if not llm_config.supports_vision:
                    screenshots = None

                model_name = llm_config.model_name

                messages = await llm_messages_builder(prompt, screenshots, llm_config.add_assistant_prefix)

                # Inject context caching system message when available
                # IMPORTANT: Only inject for extract-actions prompt to avoid contaminating other prompts
                # (e.g., check-user-goal) with the extract-action schema
                try:
                    if (
                        context
                        and context.cached_static_prompt
                        and prompt_name == EXTRACT_ACTION_PROMPT_NAME  # Only inject for extract-actions
                        and isinstance(llm_config, LLMConfig)
                        and isinstance(llm_config.model_name, str)
                    ):
                        # Check if this is an OpenAI model
                        if (
                            llm_config.model_name.startswith("gpt-")
                            or llm_config.model_name.startswith("o1-")
                            or llm_config.model_name.startswith("o3-")
                        ):
                            # For OpenAI models, we need to add the cached content as a system message
                            # and mark it for caching using the cache_control parameter
                            caching_system_message = {
                                "role": "system",
                                "content": [
                                    {
                                        "type": "text",
                                        "text": context.cached_static_prompt,
                                    }
                                ],
                            }
                            messages = [caching_system_message] + messages
                            LOG.info(
                                "Applied OpenAI context caching",
                                prompt_name=prompt_name,
                                model=llm_config.model_name,
                            )
                except Exception as e:
                    LOG.warning("Failed to apply context caching system message", error=str(e), exc_info=True)

                # Add Vertex AI cache reference only for the intended cached prompt
                vertex_cache_attached = False
                cache_resource_name = getattr(context, "vertex_cache_name", None)
                if (
                    cache_resource_name
                    and prompt_name == EXTRACT_ACTION_PROMPT_NAME
                    and getattr(context, "use_prompt_caching", False)
                    and "gemini" in model_name.lower()
                ):
                    active_parameters["cached_content"] = cache_resource_name
                    vertex_cache_attached = True
                    LOG.info(
                        "Adding Vertex AI cache reference to request",
                        prompt_name=prompt_name,
                        cache_attached=True,
                        cache_name=cache_resource_name,
                        cache_key=getattr(context, "vertex_cache_key", None),
                        cache_variant=getattr(context, "vertex_cache_variant", None),
                    )
                elif "cached_content" in active_parameters:
                    removed_cache = active_parameters.pop("cached_content", None)
                    if removed_cache:
                        LOG.info(
                            "Removed Vertex AI cache reference from request",
                            prompt_name=prompt_name,
                            cache_was_attached=True,
                            cache_name=cache_resource_name,
                            cache_key=getattr(context, "vertex_cache_key", None),
                            cache_variant=getattr(context, "vertex_cache_variant", None),
                        )

                llm_request_payload = {
                    "model": model_name,
                    "messages": messages,
                    # we're not using active_parameters here because it may contain sensitive information
                    **parameters,
                    "vertex_cache_attached": vertex_cache_attached,
                }
                llm_request_json = json.dumps(llm_request_payload)
                if should_persist_llm_artifacts:
                    artifacts.append(
                        await app.ARTIFACT_MANAGER.prepare_llm_artifact(
                            data=llm_request_json.encode("utf-8"),
                            artifact_type=ArtifactType.LLM_REQUEST,
                            **artifact_targets,
                        )
                    )

                # Strip static prompt from the request messages because it's already in the cache
                # Sending it again causes double-billing (once cached, once uncached)
                active_messages = messages
                if vertex_cache_attached and context and context.cached_static_prompt:
                    active_messages = copy.deepcopy(messages)
                    prompt_stripped = LLMAPIHandlerFactory._strip_static_prompt_from_messages(
                        active_messages, context.cached_static_prompt
                    )

                    if prompt_stripped:
                        LOG.info("Stripped static prompt from cached request to avoid double-billing")
                    else:
                        LOG.warning("Could not find static prompt to strip from cached request")

                t_llm_request = time.perf_counter()
                try:
                    # TODO (kerem): add a retry mechanism to this call (acompletion_with_retries)
                    # TODO (kerem): use litellm fallbacks? https://litellm.vercel.app/docs/tutorials/fallbacks#how-does-completion_with_fallbacks-work
                    response = await litellm.acompletion(
                        model=model_name,
                        messages=active_messages,
                        drop_params=True,  # Drop unsupported parameters gracefully
                        **active_parameters,
                    )
                except litellm.exceptions.APIError as e:
                    raise LLMProviderErrorRetryableTask(llm_key) from e
                except litellm.exceptions.ContextWindowExceededError as e:
                    duration_seconds = time.time() - start_time
                    LOG.exception(
                        "Context window exceeded",
                        llm_key=llm_key,
                        model=model_name,
                        prompt_name=prompt_name,
                        duration_seconds=duration_seconds,
                    )
                    raise SkyvernContextWindowExceededError() from e
                except CancelledError:
                    # Speculative steps are intentionally cancelled when goal verification completes first,
                    # so we log at debug level. Non-speculative cancellations are unexpected errors.
                    t_llm_cancelled = time.perf_counter()
                    if is_speculative_step:
                        LOG.debug(
                            "LLM request cancelled (speculative step)",
                            llm_key=llm_key,
                            model=model_name,
                            prompt_name=prompt_name,
                            duration=t_llm_cancelled - t_llm_request,
                        )
                        raise
                    else:
                        LOG.error(
                            "LLM request got cancelled",
                            llm_key=llm_key,
                            model=model_name,
                            prompt_name=prompt_name,
                            duration=t_llm_cancelled - t_llm_request,
                        )
                        raise LLMProviderError(llm_key) from None
                except Exception as e:
                    duration_seconds = time.time() - start_time
                    LOG.exception(
                        "LLM request failed unexpectedly",
                        llm_key=llm_key,
                        model=model_name,
                        prompt_name=prompt_name,
                        duration_seconds=duration_seconds,
                    )
                    raise LLMProviderError(llm_key) from e

                llm_response_json = _safe_model_dump_json(response)
                if should_persist_llm_artifacts:
                    artifacts.append(
                        await app.ARTIFACT_MANAGER.prepare_llm_artifact(
                            data=llm_response_json.encode("utf-8"),
                            artifact_type=ArtifactType.LLM_RESPONSE,
                            **artifact_targets,
                        )
                    )

                prompt_tokens = 0
                completion_tokens = 0
                reasoning_tokens = 0
                cached_tokens = 0
                completion_token_detail = None
                cached_token_detail = None
                try:
                    # FIXME: volcengine doesn't support litellm cost calculation.
                    llm_cost = litellm.completion_cost(completion_response=response)
                except Exception as e:
                    LOG.debug("Failed to calculate LLM cost", error=str(e), exc_info=True)
                    llm_cost = 0
                prompt_tokens = 0
                completion_tokens = 0
                reasoning_tokens = 0
                cached_tokens = 0

                if hasattr(response, "usage") and response.usage:
                    prompt_tokens = getattr(response.usage, "prompt_tokens", 0)
                    completion_tokens = getattr(response.usage, "completion_tokens", 0)

                    # Extract reasoning tokens from completion_tokens_details
                    completion_token_detail = getattr(response.usage, "completion_tokens_details", None)
                    if completion_token_detail:
                        reasoning_tokens = getattr(completion_token_detail, "reasoning_tokens", 0) or 0

                    # Extract cached tokens from prompt_tokens_details
                    cached_token_detail = getattr(response.usage, "prompt_tokens_details", None)
                    if cached_token_detail:
                        cached_tokens = getattr(cached_token_detail, "cached_tokens", 0) or 0

                    # Fallback for Vertex/Gemini: LiteLLM exposes cache_read_input_tokens on usage
                    if cached_tokens == 0:
                        cached_tokens = getattr(response.usage, "cache_read_input_tokens", 0) or 0

                _log_vertex_cache_hit_if_needed(context, prompt_name, model_name, cached_tokens)

                if step and not is_speculative_step:
                    await app.DATABASE.update_step(
                        task_id=step.task_id,
                        step_id=step.step_id,
                        organization_id=step.organization_id,
                        incremental_cost=llm_cost,
                        incremental_input_tokens=prompt_tokens if prompt_tokens > 0 else None,
                        incremental_output_tokens=completion_tokens if completion_tokens > 0 else None,
                        incremental_reasoning_tokens=reasoning_tokens if reasoning_tokens > 0 else None,
                        incremental_cached_tokens=cached_tokens if cached_tokens > 0 else None,
                    )
                if thought:
                    await app.DATABASE.update_thought(
                        thought_id=thought.observer_thought_id,
                        organization_id=thought.organization_id,
                        input_token_count=prompt_tokens if prompt_tokens > 0 else None,
                        output_token_count=completion_tokens if completion_tokens > 0 else None,
                        reasoning_token_count=reasoning_tokens if reasoning_tokens > 0 else None,
                        cached_token_count=cached_tokens if cached_tokens > 0 else None,
                        thought_cost=llm_cost,
                    )
                parsed_response = parse_api_response(response, llm_config.add_assistant_prefix, force_dict)
                parsed_response_json = json.dumps(parsed_response, indent=2)
                if should_persist_llm_artifacts:
                    artifacts.append(
                        await app.ARTIFACT_MANAGER.prepare_llm_artifact(
                            data=parsed_response_json.encode("utf-8"),
                            artifact_type=ArtifactType.LLM_RESPONSE_PARSED,
                            **artifact_targets,
                        )
                    )

                rendered_response_json = None
                if context and len(context.hashed_href_map) > 0:
                    llm_content = json.dumps(parsed_response)
                    rendered_content = Template(llm_content).render(context.hashed_href_map)
                    parsed_response = json.loads(rendered_content)
                    rendered_response_json = json.dumps(parsed_response, indent=2)
                    if should_persist_llm_artifacts:
                        artifacts.append(
                            await app.ARTIFACT_MANAGER.prepare_llm_artifact(
                                data=rendered_response_json.encode("utf-8"),
                                artifact_type=ArtifactType.LLM_RESPONSE_RENDERED,
                                **artifact_targets,
                            )
                        )

                # Track LLM API handler duration, token counts, and cost
                organization_id = organization_id or (
                    step.organization_id if step else (thought.organization_id if thought else None)
                )
                duration_seconds = time.time() - start_time
                LOG.info(
                    "LLM API handler duration metrics",
                    llm_key=llm_key,
                    prompt_name=prompt_name,
                    model=llm_config.model_name,
                    duration_seconds=duration_seconds,
                    step_id=step.step_id if step else None,
                    thought_id=thought.observer_thought_id if thought else None,
                    organization_id=organization_id,
                    input_tokens=prompt_tokens if prompt_tokens > 0 else None,
                    output_tokens=completion_tokens if completion_tokens > 0 else None,
                    reasoning_tokens=reasoning_tokens if reasoning_tokens > 0 else None,
                    cached_tokens=cached_tokens if cached_tokens > 0 else None,
                    llm_cost=llm_cost if llm_cost > 0 else None,
                )

                if step and is_speculative_step:
                    step.speculative_llm_metadata = SpeculativeLLMMetadata(
                        prompt=llm_prompt_value,
                        llm_request_json=llm_request_json,
                        llm_response_json=llm_response_json,
                        parsed_response_json=parsed_response_json,
                        rendered_response_json=rendered_response_json,
                        llm_key=llm_key,
                        model=llm_config.model_name,
                        duration_seconds=duration_seconds,
                        input_tokens=prompt_tokens if prompt_tokens > 0 else None,
                        output_tokens=completion_tokens if completion_tokens > 0 else None,
                        reasoning_tokens=reasoning_tokens if reasoning_tokens > 0 else None,
                        cached_tokens=cached_tokens if cached_tokens > 0 else None,
                        llm_cost=llm_cost if llm_cost > 0 else None,
                    )

                return parsed_response
            finally:
                try:
                    await app.ARTIFACT_MANAGER.bulk_create_artifacts(artifacts)
                except Exception:
                    LOG.error("Failed to persist artifacts", exc_info=True)

        llm_api_handler.llm_key = llm_key  # type: ignore[attr-defined]
        return llm_api_handler

    @staticmethod
    def get_api_parameters(llm_config: LLMConfig | LLMRouterConfig) -> dict[str, Any]:
        params: dict[str, Any] = {}
        if not llm_config.model_name.startswith("ollama/"):
            # OLLAMA does not support max_completion_tokens
            if llm_config.max_completion_tokens is not None:
                params["max_completion_tokens"] = llm_config.max_completion_tokens
            elif llm_config.max_tokens is not None:
                params["max_tokens"] = llm_config.max_tokens

        if llm_config.temperature is not None:
            params["temperature"] = llm_config.temperature

        if llm_config.reasoning_effort is not None:
            params["reasoning_effort"] = llm_config.reasoning_effort

        return params

    @classmethod
    def register_custom_handler(cls, llm_key: str, handler: LLMAPIHandler) -> None:
        if llm_key in cls._custom_handlers:
            raise DuplicateCustomLLMProviderError(llm_key)
        cls._custom_handlers[llm_key] = handler

    @classmethod
    def set_thinking_budget_settings(cls, settings: dict[str, int] | None) -> None:
        """Set thinking budget optimization settings for the current task/workflow."""
        cls._thinking_budget_settings = settings
        if settings:
            LOG.info("Thinking budget optimization settings applied", settings=settings)

    @classmethod
    def set_prompt_caching_settings(cls, settings: dict[str, bool] | None) -> None:
        """Set prompt caching optimization settings for the current task/workflow."""
        cls._prompt_caching_settings = settings
        if settings:
            LOG.info("Prompt caching optimization settings applied", settings=settings)


class LLMCaller:
    """
    An LLMCaller instance defines the LLM configs and keeps the chat history if needed.

    A couple of things to keep in mind:
    - LLMCaller should be compatible with litellm interface
    - LLMCaller should also support models that are not supported by litellm
    """

    def __init__(
        self,
        llm_key: str,
        screenshot_scaling_enabled: bool = False,
        base_parameters: dict[str, Any] | None = None,
    ):
        self.original_llm_key = llm_key
        self.llm_key = llm_key
        self.llm_config = LLMConfigRegistry.get_config(llm_key)
        self.base_parameters = base_parameters
        self.message_history: list[dict[str, Any]] = []
        self.current_tool_results: list[dict[str, Any]] = []
        self.screenshot_scaling_enabled = screenshot_scaling_enabled
        self.browser_window_dimension = Resolution(width=settings.BROWSER_WIDTH, height=settings.BROWSER_HEIGHT)
        self.screenshot_resize_target_dimension = self.browser_window_dimension
        if screenshot_scaling_enabled:
            self.screenshot_resize_target_dimension = get_resize_target_dimension(self.browser_window_dimension)

        self.openai_client = None
        if self.llm_key.startswith("openrouter/"):
            self.llm_key = self.llm_key.replace("openrouter/", "")
            self.openai_client = AsyncOpenAI(
                api_key=settings.OPENROUTER_API_KEY,
                base_url=settings.OPENROUTER_API_BASE,
                http_client=ForgeAsyncHttpxClientWrapper(),
            )
        elif self.llm_key == "OPENAI_COMPATIBLE" and LLMAPIHandlerFactory.is_github_copilot_endpoint():
            # For GitHub Copilot, use the actual model name from OPENAI_COMPATIBLE_MODEL_NAME
            self.llm_key = settings.OPENAI_COMPATIBLE_MODEL_NAME or self.llm_key
            self.openai_client = AsyncOpenAI(
                api_key=settings.OPENAI_COMPATIBLE_API_KEY,
                base_url=settings.OPENAI_COMPATIBLE_API_BASE,
                http_client=ForgeAsyncHttpxClientWrapper(),
            )

    def add_tool_result(self, tool_result: dict[str, Any]) -> None:
        self.current_tool_results.append(tool_result)

    def clear_tool_results(self) -> None:
        self.current_tool_results = []

    async def call(
        self,
        prompt: str | None = None,
        prompt_name: str | None = None,
        step: Step | None = None,
        task_v2: TaskV2 | None = None,
        thought: Thought | None = None,
        ai_suggestion: AISuggestion | None = None,
        screenshots: list[bytes] | None = None,
        parameters: dict[str, Any] | None = None,
        organization_id: str | None = None,
        tools: list | None = None,
        use_message_history: bool = False,
        raw_response: bool = False,
        window_dimension: Resolution | None = None,
        force_dict: bool = True,
        **extra_parameters: Any,
    ) -> dict[str, Any] | Any:
        start_time = time.perf_counter()
        active_parameters = self.base_parameters or {}
        if parameters is None:
            parameters = LLMAPIHandlerFactory.get_api_parameters(self.llm_config)

        active_parameters.update(parameters)
        if extra_parameters:
            active_parameters.update(extra_parameters)
        if self.llm_config.litellm_params:  # type: ignore
            active_parameters.update(self.llm_config.litellm_params)  # type: ignore

        context = skyvern_context.current()
        is_speculative_step = step.is_speculative if step else False
        should_persist_llm_artifacts, artifact_targets = _get_artifact_targets_and_persist_flag(
            step, is_speculative_step, task_v2, thought, ai_suggestion
        )

        artifacts: list[BulkArtifactCreationRequest | None] = []
        try:
            await _log_hashed_href_map_artifacts_if_needed(
                artifacts,
                context,
                step,
                task_v2,
                thought,
                ai_suggestion,
                is_speculative_step=is_speculative_step,
            )

            if screenshots and self.screenshot_scaling_enabled:
                target_dimension = self.get_screenshot_resize_target_dimension(window_dimension)
                if window_dimension and window_dimension != self.browser_window_dimension and tools:
                    # THIS situation only applies to Anthropic CUA
                    LOG.info(
                        "Window dimension is different from the default browser window dimension when making LLM call",
                        window_dimension=window_dimension,
                        browser_window_dimension=self.browser_window_dimension,
                    )
                    # update the tools to use the new target dimension
                    for tool in tools:
                        if "display_height_px" in tool:
                            tool["display_height_px"] = target_dimension["height"]
                        if "display_width_px" in tool:
                            tool["display_width_px"] = target_dimension["width"]
                screenshots = resize_screenshots(screenshots, target_dimension)

            llm_prompt_value = prompt or ""
            if prompt and should_persist_llm_artifacts:
                artifacts.append(
                    await app.ARTIFACT_MANAGER.prepare_llm_artifact(
                        data=prompt.encode("utf-8"),
                        artifact_type=ArtifactType.LLM_PROMPT,
                        screenshots=screenshots,
                        **artifact_targets,
                    )
                )

            if not self.llm_config.supports_vision:
                screenshots = None

            message_pattern = "openai"
            if "ANTHROPIC" in self.llm_key:
                message_pattern = "anthropic"

            if use_message_history:
                # self.message_history will be updated in place
                messages = await llm_messages_builder_with_history(
                    prompt,
                    screenshots,
                    self.message_history,
                    message_pattern=message_pattern,
                )
            else:
                messages = await llm_messages_builder_with_history(
                    prompt,
                    screenshots,
                    message_pattern=message_pattern,
                )
            llm_request_payload = {
                "model": self.llm_config.model_name,
                "messages": messages,
                # we're not using active_parameters here because it may contain sensitive information
                **parameters,
            }
            llm_request_json = json.dumps(llm_request_payload)
            if should_persist_llm_artifacts:
                artifacts.append(
                    await app.ARTIFACT_MANAGER.prepare_llm_artifact(
                        data=llm_request_json.encode("utf-8"),
                        artifact_type=ArtifactType.LLM_REQUEST,
                        **artifact_targets,
                    )
                )

            t_llm_request = time.perf_counter()
            try:
                response = await self._dispatch_llm_call(
                    messages=messages,
                    tools=tools,
                    timeout=settings.LLM_CONFIG_TIMEOUT,
                    **active_parameters,
                )
                if use_message_history:
                    # only update message_history when the request is successful
                    self.message_history = messages
            except litellm.exceptions.APIError as e:
                raise LLMProviderErrorRetryableTask(self.llm_key) from e
            except litellm.exceptions.ContextWindowExceededError as e:
                LOG.exception(
                    "Context window exceeded",
                    llm_key=self.llm_key,
                    model=self.llm_config.model_name,
                )
                raise SkyvernContextWindowExceededError() from e
            except CancelledError:
                # Speculative steps are intentionally cancelled when goal verification returns completed,
                # so we log at debug level. Non-speculative cancellations are unexpected errors.
                t_llm_cancelled = time.perf_counter()
                if is_speculative_step:
                    LOG.debug(
                        "LLM request cancelled (speculative step)",
                        llm_key=self.llm_key,
                        model=self.llm_config.model_name,
                        duration=t_llm_cancelled - t_llm_request,
                    )
                    raise
                else:
                    LOG.error(
                        "LLM request got cancelled",
                        llm_key=self.llm_key,
                        model=self.llm_config.model_name,
                        duration=t_llm_cancelled - t_llm_request,
                    )
                    raise LLMProviderError(self.llm_key) from None
            except Exception as e:
                LOG.exception("LLM request failed unexpectedly", llm_key=self.llm_key)
                raise LLMProviderError(self.llm_key) from e

            llm_response_json = _safe_model_dump_json(response)
            if should_persist_llm_artifacts:
                artifacts.append(
                    await app.ARTIFACT_MANAGER.prepare_llm_artifact(
                        data=llm_response_json.encode("utf-8"),
                        artifact_type=ArtifactType.LLM_RESPONSE,
                        **artifact_targets,
                    )
                )

            call_stats = await self.get_call_stats(response)
            if step and not is_speculative_step:
                await app.DATABASE.update_step(
                    task_id=step.task_id,
                    step_id=step.step_id,
                    organization_id=step.organization_id,
                    incremental_cost=call_stats.llm_cost,
                    incremental_input_tokens=call_stats.input_tokens,
                    incremental_output_tokens=call_stats.output_tokens,
                    incremental_reasoning_tokens=call_stats.reasoning_tokens,
                    incremental_cached_tokens=call_stats.cached_tokens,
                )
            if thought:
                await app.DATABASE.update_thought(
                    thought_id=thought.observer_thought_id,
                    organization_id=thought.organization_id,
                    input_token_count=call_stats.input_tokens,
                    output_token_count=call_stats.output_tokens,
                    reasoning_token_count=call_stats.reasoning_tokens,
                    cached_token_count=call_stats.cached_tokens,
                    thought_cost=call_stats.llm_cost,
                )

            organization_id = organization_id or (
                step.organization_id if step else (thought.organization_id if thought else None)
            )
            # Track LLM API handler duration, token counts, and cost
            duration_seconds = time.perf_counter() - start_time
            LOG.info(
                "LLM API handler duration metrics",
                llm_key=self.llm_key,
                prompt_name=prompt_name,
                model=self.llm_config.model_name,
                duration_seconds=duration_seconds,
                step_id=step.step_id if step else None,
                thought_id=thought.observer_thought_id if thought else None,
                organization_id=organization_id,
                input_tokens=call_stats.input_tokens if call_stats and call_stats.input_tokens else None,
                output_tokens=call_stats.output_tokens if call_stats and call_stats.output_tokens else None,
                reasoning_tokens=call_stats.reasoning_tokens if call_stats and call_stats.reasoning_tokens else None,
                cached_tokens=call_stats.cached_tokens if call_stats and call_stats.cached_tokens else None,
                llm_cost=call_stats.llm_cost if call_stats and call_stats.llm_cost else None,
            )

            # Raw response is used for CUA engine LLM calls.
            if raw_response:
                return response.model_dump(exclude_none=True)

            parsed_response = parse_api_response(response, self.llm_config.add_assistant_prefix, force_dict)
            parsed_response_json = json.dumps(parsed_response, indent=2)
            if should_persist_llm_artifacts:
                artifacts.append(
                    await app.ARTIFACT_MANAGER.prepare_llm_artifact(
                        data=parsed_response_json.encode("utf-8"),
                        artifact_type=ArtifactType.LLM_RESPONSE_PARSED,
                        **artifact_targets,
                    )
                )

            rendered_response_json = None
            if context and len(context.hashed_href_map) > 0:
                llm_content = json.dumps(parsed_response)
                rendered_content = Template(llm_content).render(context.hashed_href_map)
                parsed_response = json.loads(rendered_content)
                rendered_response_json = json.dumps(parsed_response, indent=2)
                if should_persist_llm_artifacts:
                    artifacts.append(
                        await app.ARTIFACT_MANAGER.prepare_llm_artifact(
                            data=rendered_response_json.encode("utf-8"),
                            artifact_type=ArtifactType.LLM_RESPONSE_RENDERED,
                            **artifact_targets,
                        )
                    )

            if step and is_speculative_step:
                step.speculative_llm_metadata = SpeculativeLLMMetadata(
                    prompt=llm_prompt_value,
                    llm_request_json=llm_request_json,
                    llm_response_json=llm_response_json,
                    parsed_response_json=parsed_response_json,
                    rendered_response_json=rendered_response_json,
                    llm_key=self.llm_key,
                    model=self.llm_config.model_name,
                    duration_seconds=duration_seconds,
                    input_tokens=call_stats.input_tokens,
                    output_tokens=call_stats.output_tokens,
                    reasoning_tokens=call_stats.reasoning_tokens,
                    cached_tokens=call_stats.cached_tokens,
                    llm_cost=call_stats.llm_cost,
                )

            return parsed_response
        finally:
            try:
                await app.ARTIFACT_MANAGER.bulk_create_artifacts(artifacts)
            except Exception:
                LOG.error("Failed to persist artifacts", exc_info=True)

    def get_screenshot_resize_target_dimension(self, window_dimension: Resolution | None) -> Resolution:
        if window_dimension and window_dimension != self.browser_window_dimension:
            return get_resize_target_dimension(window_dimension)
        return self.screenshot_resize_target_dimension

    @traced()
    async def _dispatch_llm_call(
        self,
        messages: list[dict[str, Any]],
        tools: list | None = None,
        timeout: float = settings.LLM_CONFIG_TIMEOUT,
        **active_parameters: dict[str, Any],
    ) -> ModelResponse | CustomStreamWrapper | AnthropicMessage | UITarsResponse:
        if self.openai_client:
            # Extract OpenRouter-specific and GitHub Copilot-specific parameters
            extra_headers = {}
            if settings.SKYVERN_APP_URL:
                extra_headers["HTTP-Referer"] = settings.SKYVERN_APP_URL
                extra_headers["X-Title"] = "Skyvern"

            # Add required headers for GitHub Copilot API
            if LLMAPIHandlerFactory.is_github_copilot_endpoint():
                extra_headers["Copilot-Integration-Id"] = "copilot-chat"

                # Add vision header when there are images in the request
                has_images = any(is_image_message(msg) for msg in messages)
                # Only set the header when there are actual images in the request
                if has_images:
                    extra_headers["Copilot-Vision-Request"] = "true"

            # Filter out parameters that OpenAI client doesn't support
            openai_params = {}
            if "max_completion_tokens" in active_parameters:
                openai_params["max_completion_tokens"] = active_parameters["max_completion_tokens"]
            elif "max_tokens" in active_parameters:
                openai_params["max_tokens"] = active_parameters["max_tokens"]
            if "temperature" in active_parameters:
                openai_params["temperature"] = active_parameters["temperature"]

            completion = await self.openai_client.chat.completions.create(
                model=self.llm_key,
                messages=messages,
                extra_headers=extra_headers if extra_headers else None,
                timeout=timeout,
                **openai_params,
            )
            # Convert OpenAI ChatCompletion to litellm ModelResponse format
            # litellm.utils.convert_to_model_response_object expects a dict
            response_dict = completion.model_dump()
            return litellm.ModelResponse(**response_dict)

        if self.llm_key and "ANTHROPIC" in self.llm_key:
            return await self._call_anthropic(messages, tools, timeout, **active_parameters)

        # Route UI-TARS models to custom handler instead of LiteLLM
        if self.llm_key and "UI_TARS" in self.llm_key:
            return await self._call_ui_tars(messages, tools, timeout, **active_parameters)

        return await litellm.acompletion(
            model=self.llm_config.model_name,
            messages=messages,
            tools=tools,
            timeout=timeout,
            drop_params=True,  # Drop unsupported parameters gracefully
            **active_parameters,
        )

    async def _call_anthropic(
        self,
        messages: list[dict[str, Any]],
        tools: list | None = None,
        timeout: float = settings.LLM_CONFIG_TIMEOUT,
        **active_parameters: dict[str, Any],
    ) -> AnthropicMessage:
        max_tokens = active_parameters.get("max_completion_tokens") or active_parameters.get("max_tokens") or 4096
        model_name = self.llm_config.model_name.replace("bedrock/", "").replace("anthropic/", "")
        betas = active_parameters.get("betas", NOT_GIVEN)
        thinking = active_parameters.get("thinking", NOT_GIVEN)
        LOG.info(
            "Anthropic request",
            model_name=model_name,
            betas=betas,
            tools=tools,
            timeout=timeout,
            messages_length=len(messages),
        )
        response = await app.ANTHROPIC_CLIENT.beta.messages.create(
            max_tokens=max_tokens,
            messages=messages,
            model=model_name,
            tools=tools or NOT_GIVEN,
            timeout=timeout,
            betas=betas,
            thinking=thinking,
        )
        LOG.info(
            "Anthropic response",
            model_name=model_name,
            response=response,
            betas=betas,
            tools=tools,
            timeout=timeout,
        )
        return response

    async def _call_ui_tars(
        self,
        messages: list[dict[str, Any]],
        tools: list | None = None,
        timeout: float = settings.LLM_CONFIG_TIMEOUT,
        **active_parameters: dict[str, Any],
    ) -> UITarsResponse:
        """Custom UI-TARS API call using OpenAI client with VolcEngine endpoint."""
        max_tokens = active_parameters.get("max_completion_tokens") or active_parameters.get("max_tokens") or 400
        model_name = self.llm_config.model_name.replace("volcengine/", "")

        if not app.UI_TARS_CLIENT:
            raise ValueError(
                "UI_TARS_CLIENT not initialized. Please ensure ENABLE_VOLCENGINE=true and VOLCENGINE_API_KEY is set."
            )

        LOG.info(
            "UI-TARS request",
            model_name=model_name,
            timeout=timeout,
            messages_length=len(messages),
        )

        # Use the UI-TARS client (which is OpenAI-compatible with VolcEngine)
        chat_completion: AsyncIterator[ChatCompletionChunk] = await app.UI_TARS_CLIENT.chat.completions.create(
            model=model_name,
            messages=messages,
            top_p=None,
            temperature=active_parameters.get("temperature", 0.0),
            max_tokens=max_tokens,
            stream=True,
            seed=None,
            stop=None,
            frequency_penalty=None,
            presence_penalty=None,
            timeout=timeout,
        )

        # Aggregate streaming response like in ByteDance example
        response_content = ""
        async for message in chat_completion:
            if message.choices[0].delta.content:
                response_content += message.choices[0].delta.content

        response = UITarsResponse(response_content, model_name)

        LOG.info(
            "UI-TARS response",
            model_name=model_name,
            response_length=len(response_content),
            timeout=timeout,
        )
        return response

    async def get_call_stats(
        self, response: ModelResponse | CustomStreamWrapper | AnthropicMessage | UITarsResponse
    ) -> LLMCallStats:
        empty_call_stats = LLMCallStats()
        if self.original_llm_key.startswith("openrouter/"):
            return empty_call_stats

        # Handle OPENAI_COMPATIBLE provider GitHub Copilot
        if self.original_llm_key == "OPENAI_COMPATIBLE" and isinstance(response, (ModelResponse, CustomStreamWrapper)):
            input_tokens, output_tokens, reasoning_tokens, cached_tokens = LLMAPIHandlerFactory._extract_token_counts(
                response
            )
            return LLMCallStats(
                llm_cost=0,  # TODO: calculate the cost according to the price: https://github.com/features/copilot/plans
                input_tokens=input_tokens,
                output_tokens=output_tokens,
                cached_tokens=cached_tokens,
                reasoning_tokens=reasoning_tokens,
            )

        # Handle UI-TARS response (UITarsResponse object from _call_ui_tars)
        if isinstance(response, UITarsResponse):
            ui_tars_usage = response.usage
            return LLMCallStats(
                llm_cost=0,
                # TODO: calculate the cost according to the price: https://www.volcengine.com/docs/82379/1544106
                input_tokens=ui_tars_usage.get("prompt_tokens", 0),
                output_tokens=ui_tars_usage.get("completion_tokens", 0),
                cached_tokens=0,  # only part of model support cached tokens
                reasoning_tokens=0,
            )

        if isinstance(response, AnthropicMessage):
            usage = response.usage
            input_token_cost = (3.0 / 1000000) * usage.input_tokens
            output_token_cost = (15.0 / 1000000) * usage.output_tokens
            cached_token_cost = (0.3 / 1000000) * usage.cache_read_input_tokens
            llm_cost = input_token_cost + output_token_cost + cached_token_cost
            return LLMCallStats(
                llm_cost=llm_cost,
                input_tokens=usage.input_tokens,
                output_tokens=usage.output_tokens,
                cached_tokens=usage.cache_read_input_tokens,
                reasoning_tokens=0,
            )
        elif isinstance(response, (ModelResponse, CustomStreamWrapper)):
            try:
                llm_cost = litellm.completion_cost(completion_response=response)
            except Exception as e:
                LOG.debug("Failed to calculate LLM cost", error=str(e), exc_info=True)
                llm_cost = 0
            input_tokens, output_tokens, reasoning_tokens, cached_tokens = LLMAPIHandlerFactory._extract_token_counts(
                response
            )
            return LLMCallStats(
                llm_cost=llm_cost,
                input_tokens=input_tokens,
                output_tokens=output_tokens,
                cached_tokens=cached_tokens,
                reasoning_tokens=reasoning_tokens,
            )
        return empty_call_stats


class LLMCallerManager:
    _llm_callers: dict[str, LLMCaller] = {}

    @classmethod
    def get_llm_caller(cls, uid: str) -> LLMCaller | None:
        return cls._llm_callers.get(uid)

    @classmethod
    def set_llm_caller(cls, uid: str, llm_caller: LLMCaller) -> None:
        cls._llm_callers[uid] = llm_caller

    @classmethod
    def clear_llm_caller(cls, uid: str) -> None:
        if uid in cls._llm_callers:
            del cls._llm_callers[uid]



# This file was auto-generated by Fern from our API Definition.

from __future__ import annotations

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel


class BranchConditionCriteria_Jinja2Template(UniversalBaseModel):
    criteria_type: typing.Literal["jinja2_template"] = "jinja2_template"
    expression: str
    description: typing.Optional[str] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class BranchConditionCriteria_Prompt(UniversalBaseModel):
    criteria_type: typing.Literal["prompt"] = "prompt"
    expression: str
    description: typing.Optional[str] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


BranchConditionCriteria = typing.Union[BranchConditionCriteria_Jinja2Template, BranchConditionCriteria_Prompt]



from datetime import datetime

from skyvern.forge.sdk.models import Step, StepStatus
from skyvern.forge.sdk.schemas.ai_suggestions import AISuggestion
from skyvern.forge.sdk.schemas.task_v2 import TaskV2, Thought
from skyvern.forge.sdk.schemas.workflow_runs import WorkflowRunBlock
from skyvern.schemas.workflows import BlockType

# Constants
TEST_ORGANIZATION_ID = "test-org-123"
TEST_TASK_ID = "tsk_123456789"


def create_fake_for_ai_suggestion(ai_suggestion_id: str) -> AISuggestion:
    return AISuggestion(
        ai_suggestion_id=ai_suggestion_id,
        organization_id=TEST_ORGANIZATION_ID,
        ai_suggestion_type="test_suggestion_type",
        created_at=datetime.utcnow(),
        modified_at=datetime.utcnow(),
    )


def create_fake_thought(cruise_id: str, thought_id: str) -> Thought:
    return Thought(
        observer_cruise_id=cruise_id,
        observer_thought_id=thought_id,
        created_at=datetime.utcnow(),
        modified_at=datetime.utcnow(),
        organization_id=TEST_ORGANIZATION_ID,
    )


def create_fake_step(step_id: str) -> Step:
    return Step(
        task_id=TEST_TASK_ID,
        order=1,
        retry_index=0,
        step_id=step_id,
        created_at=datetime.utcnow(),
        modified_at=datetime.utcnow(),
        status=StepStatus.created,
        is_last=False,
        organization_id=TEST_ORGANIZATION_ID,
    )


def create_fake_task_v2(observer_cruise_id: str) -> TaskV2:
    return TaskV2(
        observer_cruise_id=observer_cruise_id,
        created_at=datetime.utcnow(),
        modified_at=datetime.utcnow(),
        status=StepStatus.created,
        organization_id=TEST_ORGANIZATION_ID,
    )


def create_fake_workflow_run_block(workflow_run_id: str, workflow_run_block_id: str) -> WorkflowRunBlock:
    return WorkflowRunBlock(
        workflow_run_id=workflow_run_id,
        workflow_run_block_id=workflow_run_block_id,
        created_at=datetime.utcnow(),
        modified_at=datetime.utcnow(),
        organization_id=TEST_ORGANIZATION_ID,
        block_type=BlockType.TASK,
    )



from __future__ import annotations

import json
import sys
from typing import Any

from rich.console import Console
from rich.table import Table

console = Console()


def output(
    data: Any,
    *,
    action: str = "",
    json_mode: bool = False,
) -> None:
    if json_mode:
        envelope: dict[str, Any] = {"ok": True, "action": action, "data": data, "error": None}
        json.dump(envelope, sys.stdout, indent=2, default=str)
        sys.stdout.write("\n")
        return
    if isinstance(data, list) and data and isinstance(data[0], dict):
        table = Table()
        for key in data[0]:
            table.add_column(key.replace("_", " ").title())
        for row in data:
            table.add_row(*[str(v) for v in row.values()])
        console.print(table)
    elif isinstance(data, dict):
        for key, value in data.items():
            console.print(f"[bold]{key}:[/bold] {value}")
    else:
        console.print(str(data))


def output_error(message: str, *, hint: str = "", json_mode: bool = False, exit_code: int = 1) -> None:
    if json_mode:
        envelope: dict[str, Any] = {
            "ok": False,
            "action": "",
            "data": None,
            "error": {"message": message, "hint": hint},
        }
        json.dump(envelope, sys.stdout, indent=2, default=str)
        sys.stdout.write("\n")
        raise SystemExit(exit_code)
    console.print(f"[red]Error: {message}[/red]")
    if hint:
        console.print(f"[yellow]Hint: {hint}[/yellow]")
    raise SystemExit(exit_code)



import json
from datetime import datetime
from enum import StrEnum
from typing import Any

from pydantic import BaseModel, ConfigDict, Field, field_validator

from skyvern.forge.sdk.settings_manager import SettingsManager
from skyvern.schemas.runs import GeoTarget, ProxyLocation, ProxyLocationInput
from skyvern.utils.url_validators import validate_url

DEFAULT_WORKFLOW_TITLE = "New Workflow"


class TaskV2Status(StrEnum):
    created = "created"
    queued = "queued"
    running = "running"
    failed = "failed"
    terminated = "terminated"
    canceled = "canceled"
    timed_out = "timed_out"
    completed = "completed"

    def is_final(self) -> bool:
        return self in [self.failed, self.terminated, self.canceled, self.timed_out, self.completed]


class TaskV2(BaseModel):
    model_config = ConfigDict(from_attributes=True, populate_by_name=True)

    observer_cruise_id: str = Field(alias="task_id")
    status: TaskV2Status
    organization_id: str
    workflow_run_id: str | None = None
    workflow_id: str | None = None
    workflow_permanent_id: str | None = None
    prompt: str | None = None
    url: str | None = None
    summary: str | None = None
    output: dict[str, Any] | list | str | None = None
    totp_verification_url: str | None = None
    totp_identifier: str | None = None
    proxy_location: ProxyLocationInput = None
    webhook_callback_url: str | None = None
    webhook_failure_reason: str | None = None
    extracted_information_schema: dict | list | str | None = None
    error_code_mapping: dict | None = None
    model: dict[str, Any] | None = None
    queued_at: datetime | None = None
    started_at: datetime | None = None
    finished_at: datetime | None = None
    max_screenshot_scrolls: int | None = Field(default=None, alias="max_screenshot_scrolling_times")
    extra_http_headers: dict[str, str] | None = None
    browser_address: str | None = None
    run_with: str | None = None

    created_at: datetime
    modified_at: datetime

    @staticmethod
    def _parse_proxy_location(proxy_location: ProxyLocationInput | str) -> ProxyLocationInput:
        """Handle JSON strings that were persisted to the DB."""
        if proxy_location is None or isinstance(proxy_location, (ProxyLocation, GeoTarget, dict)):
            return proxy_location

        if isinstance(proxy_location, str):
            stripped = proxy_location.strip()
            if not stripped:
                return None

            if stripped.startswith("{"):
                try:
                    parsed = json.loads(stripped)
                    if isinstance(parsed, dict):
                        return GeoTarget.model_validate(parsed)
                except (json.JSONDecodeError, ValueError):
                    pass

            try:
                return ProxyLocation(stripped)
            except ValueError:
                return None

        return proxy_location

    @property
    def llm_key(self) -> str | None:
        """
        If the `TaskV2` has a `model` defined, then return the mapped llm_key for it.

        Otherwise return `None`.
        """

        if self.model:
            model_name = self.model.get("model_name")
            if model_name:
                mapping = SettingsManager.get_settings().get_model_name_to_llm_key()
                llm_key = mapping.get(model_name, {}).get("llm_key")
                if llm_key:
                    return llm_key

        return None

    @field_validator("url", "webhook_callback_url", "totp_verification_url")
    @classmethod
    def validate_urls(cls, url: str | None) -> str | None:
        if not url:
            return url

        return validate_url(url)

    @field_validator("proxy_location", mode="before")
    @classmethod
    def deserialize_proxy_location(cls, proxy_location: ProxyLocationInput | str) -> ProxyLocationInput:
        return cls._parse_proxy_location(proxy_location)


class ThoughtType(StrEnum):
    """
    Type of thought recorded during task execution.

    Note: Stored as VARCHAR in the database (not a PostgreSQL ENUM), so new values
    can be added without database migrations. See observer_thoughts.observer_thought_type column.
    """

    plan = "plan"
    metadata = "metadata"
    user_goal_check = "user_goal_check"
    internal_plan = "internal_plan"
    failure_describe = "failure_describe"
    termination = "termination"


class ThoughtScenario(StrEnum):
    """
    Scenario in which a thought was generated.

    Note: Stored as VARCHAR in the database (not a PostgreSQL ENUM), so new values
    can be added without database migrations. See observer_thoughts.observer_thought_scenario column.
    """

    generate_plan = "generate_plan"
    user_goal_check = "user_goal_check"
    failure_describe = "failure_describe"
    summarization = "summarization"
    generate_metadata = "generate_metadata"
    extract_loop_values = "extract_loop_values"
    generate_task_in_loop = "generate_task_in_loop"
    generate_task = "generate_general_task"
    termination = "termination"


class Thought(BaseModel):
    model_config = ConfigDict(from_attributes=True, populate_by_name=True)

    observer_thought_id: str = Field(alias="thought_id")
    observer_cruise_id: str = Field(alias="task_id")
    organization_id: str
    workflow_run_id: str | None = None
    workflow_run_block_id: str | None = None
    workflow_id: str | None = None
    workflow_permanent_id: str | None = None
    user_input: str | None = None
    observation: str | None = None
    thought: str | None = None
    answer: str | None = None
    observer_thought_type: ThoughtType | None = Field(alias="thought_type", default=ThoughtType.plan)
    observer_thought_scenario: ThoughtScenario | None = Field(alias="thought_scenario", default=None)
    output: dict[str, Any] | None = None
    input_token_count: int | None = None
    output_token_count: int | None = None
    reasoning_token_count: int | None = None
    cached_token_count: int | None = None
    thought_cost: float | None = None

    created_at: datetime
    modified_at: datetime


class TaskV2Metadata(BaseModel):
    url: str
    workflow_title: str = DEFAULT_WORKFLOW_TITLE

    @field_validator("url")
    @classmethod
    def validate_urls(cls, v: str | None) -> str | None:
        if not v:
            return v
        return validate_url(v)


class TaskV2Request(BaseModel):
    user_prompt: str
    url: str | None = None
    browser_session_id: str | None = None
    webhook_callback_url: str | None = None
    totp_verification_url: str | None = None
    totp_identifier: str | None = None
    proxy_location: ProxyLocationInput = None
    publish_workflow: bool = False
    extracted_information_schema: dict | list | str | None = None
    error_code_mapping: dict[str, str] | None = None
    max_screenshot_scrolls: int | None = None
    extra_http_headers: dict[str, str] | None = None
    browser_address: str | None = None
    run_with: str | None = None
    ai_fallback: bool = False

    @field_validator("url", "webhook_callback_url", "totp_verification_url")
    @classmethod
    def validate_urls(cls, url: str | None) -> str | None:
        if not url:
            return url

        return validate_url(url)

    @field_validator("proxy_location", mode="before")
    @classmethod
    def deserialize_proxy_location(cls, proxy_location: ProxyLocationInput | str) -> ProxyLocationInput:
        return TaskV2._parse_proxy_location(proxy_location)



"""add created_at index for multiple tables

Revision ID: e0ccabc005af
Revises: 4925c34e8d58
Create Date: 2025-09-25 15:03:25.567938+00:00

"""

from typing import Sequence, Union

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "e0ccabc005af"
down_revision: Union[str, None] = "4925c34e8d58"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index(
        "ix_observer_cruises_org_created_at", "observer_cruises", ["organization_id", "created_at"], unique=False
    )
    op.create_index(
        "ix_observer_thoughts_org_created_at", "observer_thoughts", ["organization_id", "created_at"], unique=False
    )
    op.create_index("ix_task_runs_org_created_at", "task_runs", ["organization_id", "created_at"], unique=False)
    op.create_index("ix_totp_codes_org_created_at", "totp_codes", ["organization_id", "created_at"], unique=False)
    op.create_index(
        "ix_workflow_run_blocks_org_created_at", "workflow_run_blocks", ["organization_id", "created_at"], unique=False
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_workflow_run_blocks_org_created_at", table_name="workflow_run_blocks")
    op.drop_index("ix_totp_codes_org_created_at", table_name="totp_codes")
    op.drop_index("ix_task_runs_org_created_at", table_name="task_runs")
    op.drop_index("ix_observer_thoughts_org_created_at", table_name="observer_thoughts")
    op.drop_index("ix_observer_cruises_org_created_at", table_name="observer_cruises")
    # ### end Alembic commands ###



# vendored from https://github.com/vaidik/commentjson/blob/master/commentjson/commentjson.py since that project seems to be abandoned.

import codecs
import json
import traceback
from typing import Any, TypeVar

import lark
from lark import Lark
from lark.lexer import Token
from lark.reconstruct import Reconstructor
from lark.tree import Tree

parser = Lark(
    """
    ?start: value
    ?value: object
          | array
          | string
          | SIGNED_NUMBER      -> number
          | "true"             -> true
          | "false"            -> false
          | "null"             -> null
    array  : "[" [value ("," value)*] TRAILING_COMMA? "]"
    object : "{" [pair ("," pair)*]  TRAILING_COMMA? "}"
    pair   : string ":" value
    string : ESCAPED_STRING

    COMMENT: /(#|\\/\\/)[^\\n]*/
    TRAILING_COMMA: ","

    %import common.ESCAPED_STRING
    %import common.SIGNED_NUMBER
    %import common.WS
    %ignore WS
    %ignore COMMENT
""",
    maybe_placeholders=False,
    parser="lalr",
)

serializer = Reconstructor(parser)


def detect_encoding(b: bytes) -> str:
    """
    Taken from `json` package in CPython 3.7.

    Source can be found at https://bit.ly/2OHqCIK.
    """

    bstartswith = b.startswith
    if bstartswith((codecs.BOM_UTF32_BE, codecs.BOM_UTF32_LE)):
        return "utf-32"
    if bstartswith((codecs.BOM_UTF16_BE, codecs.BOM_UTF16_LE)):
        return "utf-16"
    if bstartswith(codecs.BOM_UTF8):
        return "utf-8-sig"

    if len(b) >= 4:
        if not b[0]:
            # 00 00 -- -- - utf-32-be
            # 00 XX -- -- - utf-16-be
            return "utf-16-be" if b[1] else "utf-32-be"
        if not b[1]:
            # XX 00 00 00 - utf-32-le
            # XX 00 00 XX - utf-16-le
            # XX 00 XX -- - utf-16-le
            return "utf-16-le" if b[2] or b[3] else "utf-32-le"
    elif len(b) == 2:
        if not b[0]:
            # 00 XX - utf-16-be
            return "utf-16-be"
        if not b[1]:
            # XX 00 - utf-16-le
            return "utf-16-le"
    # default
    return "utf-8"


class BaseException(Exception):
    """Base exception to be implemented and raised while handling exceptions
    raised by libraries used in `commentjson`.

    Sets message of self in a way that it clearly calls out that the exception
    was raised by another library, along with the entire stacktrace of the
    exception raised by the other library.
    """

    library: str | None = None
    message: str

    def __init__(self, exc: Exception) -> None:
        if self.library is None:
            raise NotImplementedError("Value of library must be set in the inherited exception class.")

        tb = traceback.format_exc()
        tb = "\n".join(" " * 4 + line_ for line_ in tb.split("\n"))

        error = getattr(exc, "msg", None) or getattr(exc, "message", None) or str(exc)
        self.message = "\n".join(
            [
                "JSON Library Exception\n",
                ("Exception thrown by library ({}): \033[4;37m{}\033[0m\n".format(self.library, error)),
                "%s" % tb,
            ]
        )
        Exception.__init__(self, self.message)


class ParserException(BaseException):
    """Exception raised when the `lark` raises an exception i.e.
    the exception is not caused by `commentjson` and caused by the use of
    `lark` in `commentjson`.
    """

    library = "lark"


class JSONLibraryException(BaseException):
    """Exception raised when the `json` raises an exception i.e.
    the exception is not caused by `commentjson` and caused by the use of
    `json` in `commentjson`.

    .. note::

        As of now, ``commentjson`` supports only standard library's ``json``
        module. It might start supporting other widely-used contributed JSON
        libraries in the future.
    """

    library = "json"


T = TypeVar("T", Tree, Token)


def _remove_trailing_commas(tree: T) -> T:
    if isinstance(tree, Tree):
        tree.children = [
            _remove_trailing_commas(ch)
            for ch in tree.children
            if not (isinstance(ch, Token) and ch.type == "TRAILING_COMMA")
        ]
    return tree


def loads(text: str | bytes | bytearray, *args: Any, **kwargs: Any) -> Any:
    """Deserialize `text` (a `str` or `unicode` instance containing a JSON
    document with Python or JavaScript like comments) to a Python object.

    :param text: serialized JSON string with or without comments.
    :param kwargs: all the arguments that `json.loads <http://docs.python.org/
                   2/library/json.html#json.loads>`_ accepts.
    :returns: dict or list.
    """

    if isinstance(text, (bytes, bytearray)):
        text = text.decode(detect_encoding(text), "surrogatepass")

    try:
        parsed = _remove_trailing_commas(parser.parse(text))
        final_text = serializer.reconstruct(parsed)
    except lark.exceptions.UnexpectedCharacters:
        raise ValueError("Unable to parse text", text)

    return json.loads(final_text, *args, **kwargs)



from enum import Enum
from typing import Annotated, Any, Literal, Union

from pydantic import BaseModel, Field

from skyvern.config import settings


class SdkActionType(str, Enum):
    """Enum for SDK action types that can be executed."""

    AI_CLICK = "ai_click"
    AI_INPUT_TEXT = "ai_input_text"
    AI_SELECT_OPTION = "ai_select_option"
    AI_UPLOAD_FILE = "ai_upload_file"
    AI_ACT = "ai_act"
    EXTRACT = "extract"
    LOCATE_ELEMENT = "locate_element"
    VALIDATE = "validate"
    PROMPT = "prompt"


# Base action class
class SdkActionBase(BaseModel):
    """Base class for SDK actions."""

    type: str = Field(..., description="The type of action")

    def get_navigation_goal(self) -> str | None:
        return None

    def get_navigation_payload(self) -> dict[str, Any] | None:
        return None


# Specific action types
class ClickAction(SdkActionBase):
    """Click action parameters."""

    type: Literal["ai_click"] = "ai_click"
    selector: str | None = Field(default="", description="CSS selector for the element")
    intention: str = Field(default="", description="The intention or goal of the click")
    data: str | dict[str, Any] | None = Field(None, description="Additional context data")
    timeout: float = Field(default=settings.BROWSER_ACTION_TIMEOUT_MS, description="Timeout in milliseconds")

    def get_navigation_goal(self) -> str | None:
        return self.intention

    def get_navigation_payload(self) -> dict[str, Any] | None:
        return self.data if isinstance(self.data, dict) else None


class InputTextAction(SdkActionBase):
    """Input text action parameters."""

    type: Literal["ai_input_text"] = "ai_input_text"
    selector: str | None = Field(default="", description="CSS selector for the element")
    value: str | None = Field(default="", description="Value to input")
    intention: str = Field(default="", description="The intention or goal of the input")
    data: str | dict[str, Any] | None = Field(None, description="Additional context data")
    totp_identifier: str | None = Field(None, description="TOTP identifier for input_text actions")
    totp_url: str | None = Field(None, description="TOTP URL for input_text actions")
    timeout: float = Field(default=settings.BROWSER_ACTION_TIMEOUT_MS, description="Timeout in milliseconds")

    def get_navigation_goal(self) -> str | None:
        return self.intention

    def get_navigation_payload(self) -> dict[str, Any] | None:
        return self.data if isinstance(self.data, dict) else None


class SelectOptionAction(SdkActionBase):
    """Select option action parameters."""

    type: Literal["ai_select_option"] = "ai_select_option"
    selector: str | None = Field(default="", description="CSS selector for the element")
    value: str | None = Field(default="", description="Value to select")
    intention: str = Field(default="", description="The intention or goal of the selection")
    data: str | dict[str, Any] | None = Field(None, description="Additional context data")
    timeout: float = Field(default=settings.BROWSER_ACTION_TIMEOUT_MS, description="Timeout in milliseconds")

    def get_navigation_goal(self) -> str | None:
        return self.intention

    def get_navigation_payload(self) -> dict[str, Any] | None:
        return self.data if isinstance(self.data, dict) else None


class UploadFileAction(SdkActionBase):
    """Upload file action parameters."""

    type: Literal["ai_upload_file"] = "ai_upload_file"
    selector: str | None = Field(default="", description="CSS selector for the element")
    file_url: str | None = Field(default="", description="File URL for upload")
    intention: str = Field(default="", description="The intention or goal of the upload")
    data: str | dict[str, Any] | None = Field(None, description="Additional context data")
    timeout: float = Field(default=settings.BROWSER_ACTION_TIMEOUT_MS, description="Timeout in milliseconds")

    def get_navigation_goal(self) -> str | None:
        return self.intention

    def get_navigation_payload(self) -> dict[str, Any] | None:
        if self.data and not isinstance(self.data, dict):
            return None

        data = self.data or {}
        if "files" not in data:
            data["files"] = self.file_url
        return data


class ActAction(SdkActionBase):
    """AI act action parameters."""

    type: Literal["ai_act"] = "ai_act"
    intention: str = Field(default="", description="Natural language prompt for the action")
    data: str | dict[str, Any] | None = Field(None, description="Additional context data")

    def get_navigation_goal(self) -> str | None:
        return self.intention

    def get_navigation_payload(self) -> dict[str, Any] | None:
        return self.data if isinstance(self.data, dict) else None


class ExtractAction(SdkActionBase):
    """Extract data action parameters."""

    type: Literal["extract"] = "extract"
    prompt: str = Field(default="", description="Extraction prompt")
    extract_schema: dict[str, Any] | list | str | None = Field(None, description="Schema for extraction")
    error_code_mapping: dict[str, str] | None = Field(None, description="Error code mapping for extraction")
    intention: str | None = Field(None, description="The intention or goal of the extraction")
    data: str | dict[str, Any] | None = Field(None, description="Additional context data")

    def get_navigation_goal(self) -> str | None:
        return self.intention

    def get_navigation_payload(self) -> dict[str, Any] | None:
        return self.data if isinstance(self.data, dict) else None


class LocateElementAction(SdkActionBase):
    """Locate element action parameters."""

    type: Literal["locate_element"] = "locate_element"
    prompt: str = Field(default="", description="Natural language prompt to locate an element")

    def get_navigation_goal(self) -> str | None:
        return self.prompt

    def get_navigation_payload(self) -> dict[str, Any] | None:
        return None


class ValidateAction(SdkActionBase):
    """Validate action parameters."""

    type: Literal["validate"] = "validate"
    prompt: str = Field(..., description="Validation criteria or condition to check")
    model: dict[str, Any] | None = Field(None, description="Optional model configuration")

    def get_navigation_goal(self) -> str | None:
        return self.prompt

    def get_navigation_payload(self) -> dict[str, Any] | None:
        return None


class PromptAction(SdkActionBase):
    """Prompt action parameters."""

    type: Literal["prompt"] = "prompt"
    prompt: str = Field(..., description="The prompt to send to the LLM")
    response_schema: dict[str, Any] | None = Field(None, description="Optional JSON schema to structure the response")
    model: dict[str, Any] | None = Field(None, description="Optional model configuration")

    def get_navigation_goal(self) -> str | None:
        return self.prompt

    def get_navigation_payload(self) -> dict[str, Any] | None:
        return None


# Discriminated union of all action types
SdkAction = Annotated[
    Union[
        ClickAction,
        InputTextAction,
        SelectOptionAction,
        UploadFileAction,
        ActAction,
        ExtractAction,
        LocateElementAction,
        ValidateAction,
        PromptAction,
    ],
    Field(discriminator="type"),
]


class RunActionResponse(BaseModel):
    """Response from running an action."""

    workflow_run_id: str = Field(..., description="The workflow run ID used for this action")


class RunSdkActionRequest(BaseModel):
    """Request to run a single SDK action."""

    url: str = Field(..., description="The URL where the action should be executed")
    browser_session_id: str | None = Field(None, description="The browser session ID")
    browser_address: str | None = Field(None, description="The browser address")
    workflow_run_id: str | None = Field(
        None, description="Optional workflow run ID to continue an existing workflow run"
    )
    action: SdkAction = Field(..., description="The action to execute with its specific parameters")


class RunSdkActionResponse(BaseModel):
    """Response from running an SDK action."""

    workflow_run_id: str = Field(..., description="The workflow run ID used for this action")
    result: Any | None = Field(None, description="The result from the action (e.g., selector, value, extracted data)")



"""add 2fa waiting state fields to workflow_runs and tasks

Revision ID: a1b2c3d4e5f6
Revises: 43217e31df12
Create Date: 2026-02-13 00:00:00.000000+00:00

"""

from typing import Sequence, Union

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "a1b2c3d4e5f6"
down_revision: Union[str, None] = "43217e31df12"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.execute(
        "ALTER TABLE workflow_runs ADD COLUMN IF NOT EXISTS waiting_for_verification_code BOOLEAN NOT NULL DEFAULT false"
    )
    op.execute("ALTER TABLE workflow_runs ADD COLUMN IF NOT EXISTS verification_code_identifier VARCHAR")
    op.execute("ALTER TABLE workflow_runs ADD COLUMN IF NOT EXISTS verification_code_polling_started_at TIMESTAMP")
    op.execute(
        "ALTER TABLE tasks ADD COLUMN IF NOT EXISTS waiting_for_verification_code BOOLEAN NOT NULL DEFAULT false"
    )
    op.execute("ALTER TABLE tasks ADD COLUMN IF NOT EXISTS verification_code_identifier VARCHAR")
    op.execute("ALTER TABLE tasks ADD COLUMN IF NOT EXISTS verification_code_polling_started_at TIMESTAMP")


def downgrade() -> None:
    op.execute("ALTER TABLE tasks DROP COLUMN IF EXISTS verification_code_polling_started_at")
    op.execute("ALTER TABLE tasks DROP COLUMN IF EXISTS verification_code_identifier")
    op.execute("ALTER TABLE tasks DROP COLUMN IF EXISTS waiting_for_verification_code")
    op.execute("ALTER TABLE workflow_runs DROP COLUMN IF EXISTS verification_code_polling_started_at")
    op.execute("ALTER TABLE workflow_runs DROP COLUMN IF EXISTS verification_code_identifier")
    op.execute("ALTER TABLE workflow_runs DROP COLUMN IF EXISTS waiting_for_verification_code")



"""add run_with to workflows table and observer_cruises table

Revision ID: ac9d96ea0501
Revises: e0ccabc005af
Create Date: 2025-09-29 19:00:39.108889+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "ac9d96ea0501"
down_revision: Union[str, None] = "e0ccabc005af"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("observer_cruises", sa.Column("run_with", sa.String(), nullable=True))
    op.add_column("workflows", sa.Column("run_with", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("workflows", "run_with")
    op.drop_column("observer_cruises", "run_with")
    # ### end Alembic commands ###



"""add organization_id to script_blocks

Revision ID: a027553be970
Revises: 92d665af080e
Create Date: 2025-08-09 01:05:07.564940+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "a027553be970"
down_revision: Union[str, None] = "92d665af080e"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("script_blocks", sa.Column("organization_id", sa.String(), nullable=False))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("script_blocks", "organization_id")
    # ### end Alembic commands ###



"""add workflow sequential setting

Revision ID: 8f208daee7d1
Revises: 67add341e926
Create Date: 2025-09-18 05:29:42.472400+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "8f208daee7d1"
down_revision: Union[str, None] = "67add341e926"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("workflow_runs", sa.Column("job_id", sa.String(), nullable=True))
    op.add_column("workflows", sa.Column("run_sequentially", sa.Boolean(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("workflows", "run_sequentially")
    op.drop_column("workflow_runs", "job_id")
    # ### end Alembic commands ###



"""add_run_id_column_to_artifacts_table

Revision ID: 760ae45a1345
Revises: afeed80576cb
Create Date: 2025-06-26 14:55:09.740481+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "760ae45a1345"
down_revision: Union[str, None] = "afeed80576cb"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.add_column("artifacts", sa.Column("run_id", sa.String(), nullable=True))
    with op.get_context().autocommit_block():
        op.execute("SET statement_timeout = '3h';")
        op.execute("""
            CREATE INDEX CONCURRENTLY IF NOT EXISTS ix_artifacts_run_id
            ON artifacts (run_id)
        """)
        op.execute("RESET statement_timeout;")


def downgrade() -> None:
    op.drop_column("artifacts", "run_id")
    with op.get_context().autocommit_block():
        op.execute("SET statement_timeout = '3h';")
        op.execute("DROP INDEX CONCURRENTLY IF EXISTS ix_artifacts_run_id")
        op.execute("RESET statement_timeout;")



from enum import StrEnum
from mimetypes import add_type, guess_type
from typing import IO, Any
from urllib.parse import urlparse

import aioboto3
import structlog
from types_boto3_batch.client import BatchClient
from types_boto3_ec2.client import EC2Client
from types_boto3_ecs.client import ECSClient
from types_boto3_s3.client import S3Client
from types_boto3_secretsmanager.client import SecretsManagerClient

from skyvern.config import settings

# Register custom mime types for mimetypes guessing
add_type("application/json", ".har")
add_type("text/plain", ".log")
add_type("application/zstd", ".zst")

LOG = structlog.get_logger()


# We only include the storage classes that we want to use in our application.
class S3StorageClass(StrEnum):
    STANDARD = "STANDARD"
    # REDUCED_REDUNDANCY = "REDUCED_REDUNDANCY"
    # INTELLIGENT_TIERING = "INTELLIGENT_TIERING"
    ONEZONE_IA = "ONEZONE_IA"
    GLACIER = "GLACIER"
    GLACIER_IR = "GLACIER_IR"  # Glacier Instant Retrieval
    # DEEP_ARCHIVE = "DEEP_ARCHIVE"
    # OUTPOSTS = "OUTPOSTS"
    # STANDARD_IA = "STANDARD_IA"


class AWSClientType(StrEnum):
    S3 = "s3"
    SECRETS_MANAGER = "secretsmanager"
    ECS = "ecs"
    EC2 = "ec2"
    BATCH = "batch"


class AsyncAWSClient:
    def __init__(
        self,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        region_name: str | None = None,
        endpoint_url: str | None = None,
        profile_name: str | None = None,
    ) -> None:
        self.region_name = region_name or settings.AWS_REGION
        self._endpoint_url = endpoint_url
        self.session = aioboto3.Session(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            profile_name=profile_name,
        )

    def _ecs_client(self) -> ECSClient:
        return self.session.client(AWSClientType.ECS, region_name=self.region_name, endpoint_url=self._endpoint_url)

    def _secrets_manager_client(self) -> SecretsManagerClient:
        return self.session.client(
            AWSClientType.SECRETS_MANAGER, region_name=self.region_name, endpoint_url=self._endpoint_url
        )

    def _s3_client(self) -> S3Client:
        return self.session.client(AWSClientType.S3, region_name=self.region_name, endpoint_url=self._endpoint_url)

    def _ec2_client(self) -> EC2Client:
        return self.session.client(AWSClientType.EC2, region_name=self.region_name, endpoint_url=self._endpoint_url)

    def _batch_client(self) -> BatchClient:
        return self.session.client(AWSClientType.BATCH, region_name=self.region_name, endpoint_url=self._endpoint_url)

    def _create_tag_string(self, tags: dict[str, str]) -> str:
        return "&".join([f"{k}={v}" for k, v in tags.items()])

    async def get_secret(self, secret_name: str) -> str | None:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/secretsmanager/client/get_secret_value.html
        try:
            async with self._secrets_manager_client() as client:
                response = await client.get_secret_value(SecretId=secret_name)
                return response["SecretString"]
        except Exception as e:
            try:
                error_code = e.response["Error"]["Code"]  # type: ignore
            except Exception:
                error_code = "failed-to-get-error-code"
            LOG.exception("Failed to get secret.", secret_name=secret_name, error_code=error_code)
            return None

    async def create_secret(self, secret_name: str, secret_value: str) -> None:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/secretsmanager/client/create_secret.html
        try:
            async with self._secrets_manager_client() as client:
                await client.create_secret(Name=secret_name, SecretString=secret_value)
        except Exception as e:
            LOG.exception("Failed to create secret.", secret_name=secret_name)
            raise e

    async def set_secret(self, secret_name: str, secret_value: str) -> None:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/secretsmanager/client/put_secret_value.html
        try:
            async with self._secrets_manager_client() as client:
                await client.put_secret_value(SecretId=secret_name, SecretString=secret_value)
        except Exception as e:
            LOG.exception("Failed to set secret.", secret_name=secret_name)
            raise e

    async def delete_secret(self, secret_name: str) -> None:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/secretsmanager/client/delete_secret.html
        try:
            async with self._secrets_manager_client() as client:
                await client.delete_secret(SecretId=secret_name)
        except Exception as e:
            LOG.exception("Failed to delete secret.", secret_name=secret_name)
            raise e

    async def upload_file(
        self,
        uri: str,
        data: bytes,
        storage_class: S3StorageClass = S3StorageClass.STANDARD,
        tags: dict[str, str] | None = None,
    ) -> str | None:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3/client/put_object.html
        if storage_class not in S3StorageClass:
            raise ValueError(f"Invalid storage class: {storage_class}. Must be one of {list(S3StorageClass)}")
        try:
            async with self._s3_client() as client:
                parsed_uri = S3Uri(uri)
                extra_args = {"Tagging": self._create_tag_string(tags)} if tags else {}
                await client.put_object(
                    Body=data,
                    Bucket=parsed_uri.bucket,
                    Key=parsed_uri.key,
                    StorageClass=str(storage_class),
                    **extra_args,
                )
                return uri
        except Exception:
            LOG.exception("S3 upload failed.", uri=uri)
            return None

    async def upload_file_stream(
        self,
        uri: str,
        file_obj: IO[bytes],
        storage_class: S3StorageClass = S3StorageClass.STANDARD,
        tags: dict[str, str] | None = None,
    ) -> str | None:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3/client/upload_fileobj.html#upload-fileobj
        if storage_class not in S3StorageClass:
            raise ValueError(f"Invalid storage class: {storage_class}. Must be one of {list(S3StorageClass)}")
        try:
            async with self._s3_client() as client:
                parsed_uri = S3Uri(uri)
                extra_args: dict[str, Any] = {"StorageClass": str(storage_class)}
                if tags:
                    extra_args["Tagging"] = self._create_tag_string(tags)
                await client.upload_fileobj(
                    file_obj,
                    parsed_uri.bucket,
                    parsed_uri.key,
                    ExtraArgs=extra_args,
                )
                LOG.debug("Upload file stream success", uri=uri)
                return uri
        except Exception:
            LOG.exception("S3 upload stream failed.", uri=uri)
            return None

    async def upload_file_from_path(
        self,
        uri: str,
        file_path: str,
        storage_class: S3StorageClass = S3StorageClass.STANDARD,
        metadata: dict | None = None,
        raise_exception: bool = False,
        tags: dict[str, str] | None = None,
        content_type: str | None = None,
    ) -> None:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3/client/upload_file.html
        try:
            async with self._s3_client() as client:
                parsed_uri = S3Uri(uri)
                extra_args: dict[str, Any] = {"StorageClass": str(storage_class)}
                if metadata:
                    extra_args["Metadata"] = metadata
                if tags:
                    extra_args["Tagging"] = self._create_tag_string(tags)
                if content_type:
                    extra_args["ContentType"] = content_type
                else:
                    guessed_type, _ = guess_type(file_path)
                    if guessed_type:
                        extra_args["ContentType"] = guessed_type
                await client.upload_file(
                    Filename=file_path,
                    Bucket=parsed_uri.bucket,
                    Key=parsed_uri.key,
                    ExtraArgs=extra_args,
                )
        except Exception as e:
            LOG.exception("S3 upload failed.", uri=uri)
            if raise_exception:
                raise e

    async def download_file(self, uri: str, log_exception: bool = True) -> bytes | None:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3/client/get_object.html
        try:
            async with self._s3_client() as client:
                parsed_uri = S3Uri(uri)

                # Get full object including body
                response = await client.get_object(Bucket=parsed_uri.bucket, Key=parsed_uri.key)
                return await response["Body"].read()
        except Exception:
            if log_exception:
                LOG.exception("S3 download failed", uri=uri)
            return None

    async def delete_file(self, uri: str, log_exception: bool = True) -> None:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3/client/delete_object.html
        try:
            async with self._s3_client() as client:
                parsed_uri = S3Uri(uri)
                await client.delete_object(Bucket=parsed_uri.bucket, Key=parsed_uri.key)
        except Exception:
            if log_exception:
                LOG.exception("S3 delete failed", uri=uri)

    async def get_object_info(self, uri: str) -> dict:
        async with self._s3_client() as client:
            parsed_uri = S3Uri(uri)
            # Only get object metadata without the body
            return await client.head_object(Bucket=parsed_uri.bucket, Key=parsed_uri.key)

    async def get_file_metadata(
        self,
        uri: str,
        log_exception: bool = True,
    ) -> dict | None:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3/client/head_object.html
        """
        Retrieves only the metadata of a file without downloading its content.

        Args:
            uri: The S3 URI of the file
            client: Optional S3 client to use
            log_exception: Whether to log exceptions

        Returns:
            The metadata dictionary or None if the request fails
        """
        try:
            response = await self.get_object_info(uri)
            return response.get("Metadata", {})
        except Exception:
            if log_exception:
                LOG.exception("S3 metadata retrieval failed", uri=uri)
            return None

    async def create_presigned_urls(self, uris: list[str]) -> list[str] | None:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3/client/generate_presigned_url.html
        presigned_urls = []
        try:
            async with self._s3_client() as client:
                for uri in uris:
                    parsed_uri = S3Uri(uri)
                    url = await client.generate_presigned_url(
                        "get_object",
                        Params={"Bucket": parsed_uri.bucket, "Key": parsed_uri.key},
                        ExpiresIn=settings.PRESIGNED_URL_EXPIRATION,
                    )
                    presigned_urls.append(url)

                return presigned_urls
        except Exception:
            LOG.exception("Failed to create presigned url for S3 objects.", uris=uris)
            return None

    async def list_files(self, uri: str) -> list[str]:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3/paginator/ListObjectsV2.html
        object_keys: list[str] = []
        parsed_uri = S3Uri(uri)
        async with self._s3_client() as client:
            async for page in client.get_paginator("list_objects_v2").paginate(
                Bucket=parsed_uri.bucket, Prefix=parsed_uri.key
            ):
                if "Contents" in page:
                    for obj in page["Contents"]:
                        object_keys.append(obj["Key"])
            return object_keys

    async def delete_files(self, bucket: str, keys: list[str]) -> None:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3/client/delete_objects.html
        """
        Delete multiple objects from S3 bucket.

        Args:
            bucket: The S3 bucket name
            keys: List of object keys to delete
        """
        if not keys:
            return

        try:
            async with self._s3_client() as client:
                # Format the objects for the delete_objects call
                objects = [{"Key": key} for key in keys]

                response = await client.delete_objects(
                    Bucket=bucket,
                    Delete={
                        "Objects": objects,
                        "Quiet": False,  # Set to True to suppress response details
                    },
                )

                # Log any errors that occurred during deletion
                if "Errors" in response:
                    for error in response["Errors"]:
                        LOG.error(
                            "Failed to delete object from S3",
                            bucket=bucket,
                            key=error.get("Key"),
                            code=error.get("Code"),
                            message=error.get("Message"),
                        )
        except Exception as e:
            LOG.exception("Failed to delete files from S3", bucket=bucket, keys_count=len(keys))
            raise e

    async def restore_object(self, bucket: str, key: str, days: int = 1, tier: str = "Standard") -> None:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3/client/restore_object.html
        """
        Restore an archived S3 object from GLACIER storage class.

        Args:
            bucket: The S3 bucket name
            key: The S3 object key
            days: Number of days to keep the restored object available (default: 1)
            tier: Restoration tier - "Standard" (3-5 hours) or "Expedited" (1-5 minutes)
        """
        try:
            async with self._s3_client() as client:
                await client.restore_object(
                    Bucket=bucket, Key=key, RestoreRequest={"Days": days, "GlacierJobParameters": {"Tier": tier}}
                )
        except Exception as e:
            LOG.exception("Failed to restore S3 object", bucket=bucket, key=key, tier=tier)
            raise e

    async def run_task(
        self,
        cluster: str,
        launch_type: str,
        task_definition: str,
        subnets: list[str],
        security_groups: list[str],
        assign_public_ip: str = "DISABLED",
        enable_execute_command: bool = False,
    ) -> dict:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecs/client/run_task.html
        async with self._ecs_client() as client:
            return await client.run_task(
                cluster=cluster,
                launchType=launch_type,
                taskDefinition=task_definition,
                networkConfiguration={
                    "awsvpcConfiguration": {
                        "subnets": subnets,
                        "securityGroups": security_groups,
                        "assignPublicIp": assign_public_ip,
                    }
                },
                enableExecuteCommand=enable_execute_command,
            )

    async def stop_task(self, cluster: str, task: str, reason: str | None = None) -> dict:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecs/client/stop_task.html
        async with self._ecs_client() as client:
            return await client.stop_task(cluster=cluster, task=task, reason=reason)

    async def describe_tasks(self, cluster: str, tasks: list[str]) -> dict:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecs/client/describe_tasks.html
        async with self._ecs_client() as client:
            return await client.describe_tasks(cluster=cluster, tasks=tasks)

    async def list_tasks(self, cluster: str) -> dict:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecs/client/list_tasks.html
        async with self._ecs_client() as client:
            return await client.list_tasks(cluster=cluster)

    async def describe_task_definition(self, task_definition: str) -> dict:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecs/client/describe_task_definition.html
        async with self._ecs_client() as client:
            return await client.describe_task_definition(taskDefinition=task_definition)

    async def deregister_task_definition(self, task_definition: str) -> dict:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecs/client/deregister_task_definition.html
        async with self._ecs_client() as client:
            return await client.deregister_task_definition(taskDefinition=task_definition)

    ###### EC2 ######
    async def describe_network_interfaces(self, network_interface_ids: list[str]) -> dict:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2/client/describe_network_interfaces.html
        async with self._ec2_client() as client:
            return await client.describe_network_interfaces(NetworkInterfaceIds=network_interface_ids)

    ###### Batch ######
    async def describe_job(self, job_id: str) -> dict:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/batch/client/describe_jobs.html
        async with self._batch_client() as client:
            response = await client.describe_jobs(jobs=[job_id])
            return response["jobs"][0] if response["jobs"] else {}

    async def list_jobs(self, job_queue: str, job_status: str) -> list[dict]:
        # NOTE: AWS batch only records the latest 7 days jobs by default
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/batch/client/list_jobs.html
        async with self._batch_client() as client:
            total_jobs = []
            async for page in client.get_paginator("list_jobs").paginate(jobQueue=job_queue, jobStatus=job_status):
                for job in page["jobSummaryList"]:
                    total_jobs.append(job)

            return total_jobs

    async def submit_job(
        self,
        job_name: str,
        job_queue: str,
        job_definition: str,
        params: dict,
        job_priority: int | None = None,
        share_identifier: str | None = None,
        container_overrides: dict | None = None,
        depends_on_ids: list[str] | None = None,
    ) -> str | None:
        container_overrides = container_overrides or {}
        depends_on = [{"jobId": job_id} for job_id in depends_on_ids or []]
        async with self._batch_client() as client:
            if job_priority is None or share_identifier is None:
                response = await client.submit_job(
                    jobName=job_name,
                    jobQueue=job_queue,
                    jobDefinition=job_definition,
                    parameters=params,
                    containerOverrides=container_overrides,
                    dependsOn=depends_on,
                )
                return response.get("jobId")
            else:
                response = await client.submit_job(
                    jobName=job_name,
                    jobQueue=job_queue,
                    jobDefinition=job_definition,
                    parameters=params,
                    schedulingPriorityOverride=job_priority,
                    shareIdentifier=share_identifier,
                    containerOverrides=container_overrides,
                    dependsOn=depends_on,
                )
                return response.get("jobId")


class S3Uri:
    # From: https://stackoverflow.com/questions/42641315/s3-urls-get-bucket-name-and-path
    """
    >>> s = S3Uri("s3://bucket/hello/world")
    >>> s.bucket
    'bucket'
    >>> s.key
    'hello/world'
    >>> s.uri
    's3://bucket/hello/world'

    >>> s = S3Uri("s3://bucket/hello/world?qwe1=3#ddd")
    >>> s.bucket
    'bucket'
    >>> s.key
    'hello/world?qwe1=3#ddd'
    >>> s.uri
    's3://bucket/hello/world?qwe1=3#ddd'

    >>> s = S3Uri("s3://bucket/hello/world#foo?bar=2")
    >>> s.key
    'hello/world#foo?bar=2'
    >>> s.uri
    's3://bucket/hello/world#foo?bar=2'
    """

    def __init__(self, uri: str) -> None:
        self._parsed = urlparse(uri, allow_fragments=False)

    @property
    def bucket(self) -> str:
        return self._parsed.netloc

    @property
    def key(self) -> str:
        if self._parsed.query:
            return self._parsed.path.lstrip("/") + "?" + self._parsed.query
        else:
            return self._parsed.path.lstrip("/")

    @property
    def uri(self) -> str:
        return self._parsed.geturl()

    def __str__(self) -> str:
        return self.uri


def tag_set_to_dict(tag_set: list[dict[str, str]]) -> dict[str, str]:
    """Convert a list of tags to a dictionary."""
    return {tag["Key"]: tag["Value"] for tag in tag_set}


aws_client = AsyncAWSClient()



"""add input_or_select_context column to the actions table

Revision ID: bbcc08ba09f5
Revises: 8de03b8cb83a
Create Date: 2025-09-11 20:38:47.648188+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "bbcc08ba09f5"
down_revision: Union[str, None] = "8de03b8cb83a"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("actions", sa.Column("input_or_select_context", sa.JSON(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("actions", "input_or_select_context")
    # ### end Alembic commands ###



from typing import Any, Awaitable, Protocol

from skyvern.forge.sdk.models import Step
from skyvern.forge.sdk.schemas.ai_suggestions import AISuggestion
from skyvern.forge.sdk.schemas.task_v2 import TaskV2, Thought
from skyvern.utils.image_resizer import Resolution


class LLMAPIHandler(Protocol):
    def __call__(
        self,
        prompt: str,
        prompt_name: str,
        step: Step | None = None,
        task_v2: TaskV2 | None = None,
        thought: Thought | None = None,
        ai_suggestion: AISuggestion | None = None,
        screenshots: list[bytes] | None = None,
        parameters: dict[str, Any] | None = None,
        organization_id: str | None = None,
        tools: list | None = None,
        use_message_history: bool = False,
        raw_response: bool = False,
        window_dimension: Resolution | None = None,
        force_dict: bool = True,
    ) -> Awaitable[dict[str, Any] | Any]: ...


async def dummy_llm_api_handler(
    prompt: str,
    prompt_name: str,
    step: Step | None = None,
    task_v2: TaskV2 | None = None,
    thought: Thought | None = None,
    ai_suggestion: AISuggestion | None = None,
    screenshots: list[bytes] | None = None,
    parameters: dict[str, Any] | None = None,
    organization_id: str | None = None,
    tools: list | None = None,
    use_message_history: bool = False,
    raw_response: bool = False,
    window_dimension: Resolution | None = None,
    force_dict: bool = True,
) -> dict[str, Any] | Any:
    raise NotImplementedError("Your LLM provider is not configured. Please configure it in the .env file.")



# This file was auto-generated by Fern from our API Definition.

from __future__ import annotations

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .workflow_parameter_type import WorkflowParameterType
from .workflow_parameter_yaml_default_value import WorkflowParameterYamlDefaultValue


class WorkflowDefinitionYamlParametersItem_AwsSecret(UniversalBaseModel):
    parameter_type: typing.Literal["aws_secret"] = "aws_secret"
    key: str
    description: typing.Optional[str] = None
    aws_key: str

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class WorkflowDefinitionYamlParametersItem_AzureVaultCredential(UniversalBaseModel):
    parameter_type: typing.Literal["azure_vault_credential"] = "azure_vault_credential"
    key: str
    description: typing.Optional[str] = None
    vault_name: str
    username_key: str
    password_key: str
    totp_secret_key: typing.Optional[str] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class WorkflowDefinitionYamlParametersItem_BitwardenCreditCardData(UniversalBaseModel):
    parameter_type: typing.Literal["bitwarden_credit_card_data"] = "bitwarden_credit_card_data"
    key: str
    description: typing.Optional[str] = None
    bitwarden_client_id_aws_secret_key: str
    bitwarden_client_secret_aws_secret_key: str
    bitwarden_master_password_aws_secret_key: str
    bitwarden_collection_id: str
    bitwarden_item_id: str

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class WorkflowDefinitionYamlParametersItem_BitwardenLoginCredential(UniversalBaseModel):
    parameter_type: typing.Literal["bitwarden_login_credential"] = "bitwarden_login_credential"
    key: str
    description: typing.Optional[str] = None
    bitwarden_client_id_aws_secret_key: str
    bitwarden_client_secret_aws_secret_key: str
    bitwarden_master_password_aws_secret_key: str
    url_parameter_key: typing.Optional[str] = None
    bitwarden_collection_id: typing.Optional[str] = None
    bitwarden_item_id: typing.Optional[str] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class WorkflowDefinitionYamlParametersItem_BitwardenSensitiveInformation(UniversalBaseModel):
    parameter_type: typing.Literal["bitwarden_sensitive_information"] = "bitwarden_sensitive_information"
    key: str
    description: typing.Optional[str] = None
    bitwarden_client_id_aws_secret_key: str
    bitwarden_client_secret_aws_secret_key: str
    bitwarden_master_password_aws_secret_key: str
    bitwarden_collection_id: str
    bitwarden_identity_key: str
    bitwarden_identity_fields: typing.List[str]

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class WorkflowDefinitionYamlParametersItem_Context(UniversalBaseModel):
    parameter_type: typing.Literal["context"] = "context"
    key: str
    description: typing.Optional[str] = None
    source_parameter_key: str

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class WorkflowDefinitionYamlParametersItem_Credential(UniversalBaseModel):
    parameter_type: typing.Literal["credential"] = "credential"
    key: str
    description: typing.Optional[str] = None
    credential_id: str

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class WorkflowDefinitionYamlParametersItem_Onepassword(UniversalBaseModel):
    parameter_type: typing.Literal["onepassword"] = "onepassword"
    key: str
    description: typing.Optional[str] = None
    vault_id: str
    item_id: str

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class WorkflowDefinitionYamlParametersItem_Output(UniversalBaseModel):
    parameter_type: typing.Literal["output"] = "output"
    key: str
    description: typing.Optional[str] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class WorkflowDefinitionYamlParametersItem_Workflow(UniversalBaseModel):
    parameter_type: typing.Literal["workflow"] = "workflow"
    key: str
    description: typing.Optional[str] = None
    workflow_parameter_type: WorkflowParameterType
    default_value: typing.Optional[WorkflowParameterYamlDefaultValue] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


WorkflowDefinitionYamlParametersItem = typing.Union[
    WorkflowDefinitionYamlParametersItem_AwsSecret,
    WorkflowDefinitionYamlParametersItem_AzureVaultCredential,
    WorkflowDefinitionYamlParametersItem_BitwardenCreditCardData,
    WorkflowDefinitionYamlParametersItem_BitwardenLoginCredential,
    WorkflowDefinitionYamlParametersItem_BitwardenSensitiveInformation,
    WorkflowDefinitionYamlParametersItem_Context,
    WorkflowDefinitionYamlParametersItem_Credential,
    WorkflowDefinitionYamlParametersItem_Onepassword,
    WorkflowDefinitionYamlParametersItem_Output,
    WorkflowDefinitionYamlParametersItem_Workflow,
]



"""
Module for generating GeneratedWorkflowParameters schema from workflow run input_text actions.
"""

from typing import Any, Dict, List, Tuple

import structlog
from pydantic import BaseModel

from skyvern.forge import app
from skyvern.forge.sdk.core import skyvern_context
from skyvern.forge.sdk.prompting import PromptEngine
from skyvern.webeye.actions.actions import ActionType

LOG = structlog.get_logger(__name__)

# Initialize prompt engine
prompt_engine = PromptEngine("skyvern")
CUSTOM_FIELD_ACTIONS = [ActionType.INPUT_TEXT, ActionType.UPLOAD_FILE, ActionType.SELECT_OPTION]


class GeneratedFieldMapping(BaseModel):
    """Mapping of action indices to field names."""

    field_mappings: Dict[str, str]
    schema_fields: Dict[str, Dict[str, str]]


async def generate_workflow_parameters_schema(
    actions_by_task: Dict[str, List[Dict[str, Any]]],
    existing_field_assignments: Dict[int, str] | None = None,
) -> Tuple[str, Dict[str, str]]:
    """
    Generate a GeneratedWorkflowParameters Pydantic schema based on input_text actions.

    Args:
        actions_by_task: Dictionary mapping task IDs to lists of action dictionaries
        existing_field_assignments: Optional dictionary mapping action index (1-based) to
            existing field names that must be preserved. Used when regenerating schemas
            to maintain compatibility with cached block code.

    Returns:
        Tuple of (schema_code, field_mappings) where:
        - schema_code: Python code for the GeneratedWorkflowParameters class
        - field_mappings: Dictionary mapping action indices to field names for hydration
    """
    existing_field_assignments = existing_field_assignments or {}

    # Extract all input_text actions
    custom_field_actions = []
    action_index_map = {}
    action_counter = 1

    for task_id, actions in actions_by_task.items():
        for action in actions:
            action_type = action.get("action_type", "")
            if action_type not in CUSTOM_FIELD_ACTIONS:
                continue

            value = ""
            if action_type == ActionType.INPUT_TEXT:
                value = action.get("text", "")
            elif action_type == ActionType.UPLOAD_FILE:
                value = action.get("file_url", "")
            elif action_type == ActionType.SELECT_OPTION:
                value = action.get("option", "")

            # Skip actions without data - addresses race condition where script generation
            # runs before action data is fully saved to database (SKY-7653)
            if not value:
                LOG.debug(
                    "Skipping action without data during field mapping generation",
                    action_type=action_type,
                    action_id=action.get("action_id", ""),
                    task_id=task_id,
                )
                # Mark that we had incomplete actions - triggers finalize regeneration
                ctx = skyvern_context.current()
                if ctx:
                    ctx.script_gen_had_incomplete_actions = True
                continue

            # Check if this action has an existing field name that must be preserved
            existing_field_name = existing_field_assignments.get(action_counter)

            custom_field_actions.append(
                {
                    "action_type": action_type,
                    "value": value,
                    "intention": action.get("intention", ""),
                    "task_id": task_id,
                    "action_id": action.get("action_id", ""),
                    "existing_field_name": existing_field_name,
                }
            )
            action_index_map[f"action_index_{action_counter}"] = {
                "task_id": task_id,
                "action_id": action.get("action_id", ""),
            }
            action_counter += 1

    if not custom_field_actions:
        LOG.warning("No field_name_actions found in workflow run")
        return _generate_empty_schema(), {}

    # Generate field names using LLM
    try:
        field_mapping = await _generate_field_names_with_llm(custom_field_actions)

        # Generate the Pydantic schema code
        schema_code = _generate_pydantic_schema(field_mapping.schema_fields)

        # Create field mappings for action hydration
        action_field_mappings = {}
        for action_idx, field_name in field_mapping.field_mappings.items():
            if action_idx in action_index_map:
                action_info = action_index_map[action_idx]
                key = f"{action_info['task_id']}:{action_info['action_id']}"
                action_field_mappings[key] = field_name

        return schema_code, action_field_mappings

    except Exception as e:
        LOG.error("Failed to generate workflow parameters schema", error=str(e), exc_info=True)
        return _generate_empty_schema(), {}


async def _generate_field_names_with_llm(
    custom_field_actions: List[Dict[str, Any]],
) -> GeneratedFieldMapping:
    """
    Use LLM to generate field names from input actions.

    Args:
        custom_field_actions: List of action dictionaries with action details.
            Each action may include an "existing_field_name" key if the field
            name must be preserved from a cached block.

    Returns:
        GeneratedFieldMapping with field mappings and schema definitions
    """
    # Check if any actions have existing field names that must be preserved
    has_existing_fields = any(action.get("existing_field_name") for action in custom_field_actions)

    prompt = prompt_engine.load_prompt(
        template="generate-workflow-parameters",
        custom_field_actions=custom_field_actions,
        has_existing_fields=has_existing_fields,
    )

    response = await app.SCRIPT_GENERATION_LLM_API_HANDLER(prompt=prompt, prompt_name="generate-workflow-parameters")

    return GeneratedFieldMapping.model_validate(response)


def _generate_pydantic_schema(schema_fields: Dict[str, Dict[str, str]]) -> str:
    """
    Generate Pydantic schema code from field definitions.

    Args:
        schema_fields: Dictionary of field names to their type and description

    Returns:
        Python code string for the GeneratedWorkflowParameters class
    """
    if not schema_fields:
        return _generate_empty_schema()

    lines = [
        "class GeneratedWorkflowParameters(BaseModel):",
        '    """Generated schema representing all input_text action values from the workflow run."""',
        "",
    ]

    for field_name, field_info in schema_fields.items():
        field_type = field_info.get("type", "str")
        description = field_info.get("description", f"Value for {field_name}")

        # Escape quotes in description
        description = description.replace('"', '\\"')

        lines.append(f'    {field_name}: {field_type} = Field(description="{description}", default="")')

    return "\n".join(lines)


def _generate_empty_schema() -> str:
    """Generate an empty schema when no input_text actions are found."""
    return '''from pydantic import BaseModel


class GeneratedWorkflowParameters(BaseModel):
    """Generated schema representing all input_text action values from the workflow run."""
    pass
'''


def hydrate_input_text_actions_with_field_names(
    actions_by_task: Dict[str, List[Dict[str, Any]]], field_mappings: Dict[str, str]
) -> Dict[str, List[Dict[str, Any]]]:
    """
    Add field_name to input_text actions based on generated mappings.

    Args:
        actions_by_task: Dictionary mapping task IDs to lists of action dictionaries
        field_mappings: Dictionary mapping "task_id:action_id" to field names

    Returns:
        Updated actions_by_task with field_name added to input_text actions
    """
    updated_actions_by_task = {}

    for task_id, actions in actions_by_task.items():
        updated_actions = []

        for action in actions:
            action_copy = action.copy()

            if action.get("action_type") in CUSTOM_FIELD_ACTIONS:
                action_id = action.get("action_id", "")
                mapping_key = f"{task_id}:{action_id}"

                if mapping_key in field_mappings:
                    action_copy["field_name"] = field_mappings[mapping_key]
                # else:
                #     # Fallback field name if mapping not found
                #     intention = action.get("intention", "")
                #     if intention:
                #         # Simple field name generation from intention
                #         field_name = intention.lower().replace(" ", "_").replace("?", "").replace("'", "")
                #         field_name = "".join(c for c in field_name if c.isalnum() or c == "_")
                #         action_copy["field_name"] = field_name or "unknown_field"
                #     else:
                #         action_copy["field_name"] = "unknown_field"

            updated_actions.append(action_copy)

        updated_actions_by_task[task_id] = updated_actions

    return updated_actions_by_task
